/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: sp_wifi_addr4.c,v 1.3 2003/08/26 01:37:21 andrew Exp $ */

/* sp_wifi_addr4
 * 
 * Purpose:
 *
 * Tests 802.11 frames 4th address field.
 *
 * Arguments:
 *   
 * {{ '!' | {} } <MAC> }
 *
 * Effect:
 *
 * Success on address field match, failure otherwise.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "plugbase.h"
#include "parser.h"
#include "debug.h"
#include "util.h"
#include "plugin_enum.h"

#include "wifi_datatypes.h"

/* don't forget to include the name of this file in plugbase.h! */

extern char *file_name;  /* this is the file name from rules.c, generally used
                            for error messages */

extern int file_line;    /* this is the file line number from rules.c that is
                            used to indicate file lines for error messages */

typedef struct _WifiAddr4Data
{
  u_int8_t addr4[6];
  u_int8_t not_flag;
} WifiAddr4Data;

void WifiAddr4Init(char *, OptTreeNode *, int);
void WifiAddr4RuleParseFunction(char *, OptTreeNode *);
int WifiAddr4DetectorFunction(Packet *, struct _OptTreeNode *, OptFpList *);

/****************************************************************************
 * 
 * Function: SetupWifiAddr4()
 *
 * Purpose: Generic detection engine plugin template.  Registers the
 *          configuration function and links it to a rule keyword.  This is
 *          the function that gets called from InitPlugins in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void SetupWifiAddr4()
{
    /* map the keyword to an initialization/processing function */
    RegisterPlugin("addr4", WifiAddr4Init);

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Plugin: WifiAddr4 Setup\n"););
}


/****************************************************************************
 * 
 * Function: WifiAddr4Init(char *, OptTreeNode *)
 *
 * Purpose: Generic rule configuration function.  Handles parsing the rule 
 *          information and attaching the associated detection function to
 *          the OTN.
 *
 * Arguments: data => rule arguments/data
 *            otn => pointer to the current rule option list node
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiAddr4Init(char *data, OptTreeNode *otn, int protocol)
{
    /* allocate the data structure and attach it to the
       rule's data struct list */
    otn->ds_list[PLUGIN_WIFI_ADDR4_CHECK] = (WifiAddr4Data *) calloc(sizeof(WifiAddr4Data), sizeof(char));

    /* be sure to check that the protocol that is passed in matches the
       transport layer protocol that you're using for this rule! */

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    WifiAddr4RuleParseFunction(data, otn);

    /* finally, attach the option's detection function to the rule's 
       detect function pointer list */
    AddOptFuncToList(WifiAddr4DetectorFunction, otn);
}



/****************************************************************************
 * 
 * Function: WifiAddr4RuleParseFunction(char *, OptTreeNode *)
 *
 * Purpose: This is the function that is used to process the option keyword's
 *          arguments and attach them to the rule's data structures.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiAddr4RuleParseFunction(char *data, OptTreeNode *otn)
{
    WifiAddr4Data *ds_ptr;  /* data struct pointer */
	 
    /* set the ds pointer to make it easier to reference the option's
       particular data struct */
    ds_ptr = otn->ds_list[PLUGIN_WIFI_ADDR4_CHECK];

    /* manipulate the option arguments here */
    while(isspace((int) *data)) data++;
    
    if(*data == '!'){
      ds_ptr->not_flag = 1;
      data++;
    }

    /* set the final option arguments here */
    
    if(WifiAddr4ParseMAC(ds_ptr->addr4, data) < 0)    
      FatalError("%s (%d) => Error parsing MAC address \"%s\"\n", file_name, file_line, data);
}



/*
 * Function: WifiAddr4ParseMAC(u_int8_t *, char *)
 *
 * Purpose: Parses BSSIDs from a character string and stores them
 *          in a byte array.
 *
 * Arguments: bssid => pointer to byte array to store parsed BSSID in
 *            mac_str => character string containing BSSID
 *
 * Returns: 0 on success, -1 on error
 *
 */

int WifiAddr4ParseMAC(u_int8_t *bssid, char *mac_str){
  
  u_int64_t mac;
  
  if(strlen(mac_str) != 14)
	 return -1;
  
  mac = strtoll(mac_str, NULL, 16);
  if(mac < 0 || mac > 0xffffffffffff)
	 return -1;
  
  bssid[0] = mac >> 40 & 0x00ff;
  bssid[1] = mac >> 32 & 0x00ff;
  bssid[2] = mac >> 24 & 0x00ff;
  bssid[3] = mac >> 16 & 0x00ff;
  bssid[4] = mac >> 8 & 0x00ff;
  bssid[5] = mac & 0x00ff;
  
  return 0;
}



/****************************************************************************
 * 
 * Function: WifiAddr4DetectorFunction(char *, OptTreeNode *)
 *
 * Purpose: Use this function to perform the particular detection routine
 *          that this rule keyword is supposed to encompass.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: If the detection test fails, this function *must* return a zero!
 *          On success, it calls the next function in the detection list 
 *
 ****************************************************************************/
int WifiAddr4DetectorFunction(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
  
  WifiAddr4Data *plugin_data;
  plugin_data = otn->ds_list[PLUGIN_WIFI_ADDR4_CHECK];

  if(!p->wifih){
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Not a WIFI frame\n"););
    return 0;
  }

  if(p->wifih->type != TYPE_DATA || p->wifih->to_ds != 1 || p->wifih->from_ds != 1){
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Frame does not contain addr4 field\n"););
	 return 0;
  }

  /* your detection function tests go here */
  if((plugin_data->not_flag && (memcmp(p->wifih->addr3, plugin_data->addr4, 6) != 0)) || 
	  (!plugin_data->not_flag && (memcmp(p->wifih->addr3, plugin_data->addr4, 6) == 0)))
	 {
		return fp_list->next->OptTestFunc(p, otn, fp_list->next);
	 }
#ifdef DEBUG
  else
	 {
		/* you can put debug comments here or not */
		DebugMessage(DEBUG_PLUGIN,"No match\n");
	 }
#endif

  /* if the test isn't successful, this function *must* return 0 */
  return 0;
}



