/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: sp_wifi_bssid.c,v 1.6 2003/08/24 06:45:32 andrew Exp $ */

/* sp_wifi_bssid
 * 
 * Purpose:
 *
 * Tests 802.11 frames BSSID.
 *
 * Arguments:
 *   
 * {{ '!' | {} } <MAC> }
 *
 * Effect:
 *
 * Success on BSSID match, failure otherwise.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "plugbase.h"
#include "parser.h"
#include "debug.h"
#include "util.h"
#include "plugin_enum.h"

#include "wifi_datatypes.h"

/* don't forget to include the name of this file in plugbase.h! */

extern char *file_name;  /* this is the file name from rules.c, generally used
                            for error messages */

extern int file_line;    /* this is the file line number from rules.c that is
                            used to indicate file lines for error messages */

typedef struct _WifiBssidData
{
  u_int8_t bssid[6];
  u_int8_t not_flag;
} WifiBssidData;

void WifiBssidInit(char *, OptTreeNode *, int);
void WifiBssidRuleParseFunction(char *, OptTreeNode *);
int WifiBssidDetectorFunction(Packet *, struct _OptTreeNode *, OptFpList *);

/****************************************************************************
 * 
 * Function: SetupWifiBssid()
 *
 * Purpose: Generic detection engine plugin template.  Registers the
 *          configuration function and links it to a rule keyword.  This is
 *          the function that gets called from InitPlugins in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void SetupWifiBssid()
{
    /* map the keyword to an initialization/processing function */
    RegisterPlugin("bssid", WifiBssidInit);

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Plugin: WifiBssid Setup\n"););
}


/****************************************************************************
 * 
 * Function: WifiBssidInit(char *, OptTreeNode *)
 *
 * Purpose: Generic rule configuration function.  Handles parsing the rule 
 *          information and attaching the associated detection function to
 *          the OTN.
 *
 * Arguments: data => rule arguments/data
 *            otn => pointer to the current rule option list node
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiBssidInit(char *data, OptTreeNode *otn, int protocol)
{
    /* allocate the data structure and attach it to the
       rule's data struct list */
    otn->ds_list[PLUGIN_WIFI_BSSID_CHECK] = (WifiBssidData *) calloc(sizeof(WifiBssidData), sizeof(char));

    /* be sure to check that the protocol that is passed in matches the
       transport layer protocol that you're using for this rule! */

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    WifiBssidRuleParseFunction(data, otn);

    /* finally, attach the option's detection function to the rule's 
       detect function pointer list */
    AddOptFuncToList(WifiBssidDetectorFunction, otn);
}



/****************************************************************************
 * 
 * Function: WifiBssidRuleParseFunction(char *, OptTreeNode *)
 *
 * Purpose: This is the function that is used to process the option keyword's
 *          arguments and attach them to the rule's data structures.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiBssidRuleParseFunction(char *data, OptTreeNode *otn)
{
    WifiBssidData *ds_ptr;  /* data struct pointer */
	 
    /* set the ds pointer to make it easier to reference the option's
       particular data struct */
    ds_ptr = otn->ds_list[PLUGIN_WIFI_BSSID_CHECK];

    /* manipulate the option arguments here */
    while(isspace((int) *data)) data++;
    
    if(*data == '!'){
      ds_ptr->not_flag = 1;
      data++;
    }

    /* set the final option arguments here */
    
    if(WifiBssidParseBSSID(ds_ptr->bssid, data) < 0)    
      FatalError("%s (%d) => Error parsing BSSID \"%s\"\n", file_name, file_line, data);
}



/*
 * Function: WifiBssidParseBSSID(u_int8_t *, char *)
 *
 * Purpose: Parses BSSIDs from a character string and stores them
 *          in a byte array.
 *
 * Arguments: bssid => pointer to byte array to store parsed BSSID in
 *            mac_str => character string containing BSSID
 *
 * Returns: 0 on success, -1 on error
 *
 */

int WifiBssidParseBSSID(u_int8_t *bssid, char *mac_str){

  u_int64_t mac;
 
  if(strlen(mac_str) != 14)
	 return -1;
  
  mac = strtoll(mac_str, NULL, 16);
  if(mac < 0 || mac > 0xffffffffffff)
	 return -1;
  
  bssid[0] = mac >> 40 & 0x00ff;
  bssid[1] = mac >> 32 & 0x00ff;
  bssid[2] = mac >> 24 & 0x00ff;
  bssid[3] = mac >> 16 & 0x00ff;
  bssid[4] = mac >> 8 & 0x00ff;
  bssid[5] = mac & 0x00ff;

  return 0;
}



/****************************************************************************
 * 
 * Function: WifiBssidDetectorFunction(char *, OptTreeNode *)
 *
 * Purpose: Use this function to perform the particular detection routine
 *          that this rule keyword is supposed to encompass.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: If the detection test fails, this function *must* return a zero!
 *          On success, it calls the next function in the detection list 
 *
 ****************************************************************************/
int WifiBssidDetectorFunction(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
  
  WifiBssidData *plugin_data;
  u_int8_t *bssid;
  plugin_data = otn->ds_list[PLUGIN_WIFI_BSSID_CHECK];

  if(!p->wifih){
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Not a WIFI frame\n"););
    return 0;
  }

  switch(p->wifih->type){		  /* point bssid to the correct field in the frame */
	 
  case TYPE_MANAGEMENT:
	 bssid = p->wifih->addr3;
	 break;

  case TYPE_CONTROL:
	 if(p->wifih->stype == STYPE_RTS || p->wifih->stype == STYPE_CTS || p->wifih->stype == STYPE_ACK){
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Received Control frame subtype that does not contain BSSID\n"););
		return 0;
	 }
	 else if(p->wifih->stype == STYPE_PS)
		bssid = p->wifih->addr1;
	 else if(p->wifih->stype == STYPE_CFEND)
		bssid = p->wifih->addr2;
	 else if(p->wifih->stype == STYPE_CFEND_CFACK)
		bssid = p->wifih->addr2;
	 else{
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Received Control frame with unknown/invalid subtype\n"););
		return 0;
	 }
	 break;
	 
  case TYPE_DATA:
	 if(p->wifih->to_ds){		
		if(p->wifih->from_ds){	  /* to_ds == 1 && from_ds == 1 */
		  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Received Data frame subtype that does not contain BSSID\n"););
		  return 0;
		}
		else							  /* to_ds == 1 && from_ds == 0 */
		  bssid = p->wifih->addr1;
	 }
	 else{
		if(p->wifih->from_ds)	  /* to_ds == 0 && from_ds == 1 */
		  bssid = p->wifih->addr2;
		else							  /* to_ds == 0 && from_ds == 0 */
		  bssid = p->wifih->addr3;
	 }
	 break;
	 
  default:
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Received frame with unknown/invalid type\n"););
	 return 0;
  }


  /* your detection function tests go here */
  if((plugin_data->not_flag && (memcmp(bssid, plugin_data->bssid, 6) != 0)) || 
	  (!plugin_data->not_flag && (memcmp(bssid, plugin_data->bssid, 6) == 0)))
	 {
		return fp_list->next->OptTestFunc(p, otn, fp_list->next);
	 }
#ifdef DEBUG
  else
	 {
		/* you can put debug comments here or not */
		DebugMessage(DEBUG_PLUGIN,"No match\n");
	 }
#endif

  /* if the test isn't successful, this function *must* return 0 */
  return 0;
}



