/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: sp_wifi_frame_control.c,v 1.4 2003/07/27 01:32:50 andrew Exp $ */

/* sp_wifi_frame_control
 * 
 * Purpose:
 *
 * Tests 802.11 frames frame_control field.
 *
 * Arguments:
 *   
 * {{ '!' | {} } { 0 <= n <= 65535 } }
 * 
 * n may be specified in either hex or decimal
 *
 * Effect:
 *
 * Success on frame_control field match, failure otherwise.
 *
 */

#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "plugbase.h"
#include "parser.h"
#include "debug.h"
#include "util.h"
#include "plugin_enum.h"


/* don't forget to include the name of this file in plugbase.h! */

extern char *file_name;  /* this is the file name from rules.c, generally used
                            for error messages */

extern int file_line;    /* this is the file line number from rules.c that is
                            used to indicate file lines for error messages */

typedef struct _WifiFrameControlData
{
  u_int16_t frame_control;
  u_int8_t not_flag;
} WifiFrameControlData;

void WifiFrameControlInit(char *, OptTreeNode *, int);
void WifiFrameControlRuleParseFunction(char *, OptTreeNode *);
int WifiFrameControlDetectorFunction(Packet *, struct _OptTreeNode *, OptFpList *);

/****************************************************************************
 * 
 * Function: SetupWifiFrameControl()
 *
 * Purpose: Generic detection engine plugin template.  Registers the
 *          configuration function and links it to a rule keyword.  This is
 *          the function that gets called from InitPlugins in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void SetupWifiFrameControl()
{
    /* map the keyword to an initialization/processing function */
    RegisterPlugin("frame_control", WifiFrameControlInit);

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Plugin: WifiFrameControl Setup\n"););
}


/****************************************************************************
 * 
 * Function: WifiFrameControlInit(char *, OptTreeNode *)
 *
 * Purpose: Generic rule configuration function.  Handles parsing the rule 
 *          information and attaching the associated detection function to
 *          the OTN.
 *
 * Arguments: data => rule arguments/data
 *            otn => pointer to the current rule option list node
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiFrameControlInit(char *data, OptTreeNode *otn, int protocol)
{
    /* allocate the data structure and attach it to the
       rule's data struct list */
    otn->ds_list[PLUGIN_WIFI_FRAME_CONTROL_CHECK] = (WifiFrameControlData *) calloc(sizeof(WifiFrameControlData), sizeof(char));

    /* be sure to check that the protocol that is passed in matches the
       transport layer protocol that you're using for this rule! */

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    WifiFrameControlRuleParseFunction(data, otn);

    /* finally, attach the option's detection function to the rule's 
       detect function pointer list */
    AddOptFuncToList(WifiFrameControlDetectorFunction, otn);
}



/****************************************************************************
 * 
 * Function: WifiFrameControlRuleParseFunction(char *, OptTreeNode *)
 *
 * Purpose: This is the function that is used to process the option keyword's
 *          arguments and attach them to the rule's data structures.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiFrameControlRuleParseFunction(char *data, OptTreeNode *otn)
{
    WifiFrameControlData *ds_ptr;  /* data struct pointer */
	 unsigned long val = 1 << 16;

    /* set the ds pointer to make it easier to reference the option's
       particular data struct */
    ds_ptr = otn->ds_list[PLUGIN_WIFI_FRAME_CONTROL_CHECK];

    /* manipulate the option arguments here */
    while(isspace((int) *data)) data++;
    
    if(*data == '!'){
      ds_ptr->not_flag = 1;
      data++;
    }

    /* set the final option arguments here */
	 
	 if(*data)
		if(strncmp(data, "0x", 2) == 0)
		  val = strtoul(data, NULL, 16); /* use strtoul so people don't have to do
														sign-extended 2's complement hex*/
		else 
		  val = strtol(data, NULL, 10);

	 
	 if(val < 0 || val > 65535)
		FatalError("%s(%d) => frame_control value out of range \"%s\"\n",
					  file_name, file_line, data);
	 
	 ds_ptr->frame_control = val;
}


/****************************************************************************
 * 
 * Function: WifiFrameControlDetectorFunction(char *, OptTreeNode *)
 *
 * Purpose: Use this function to perform the particular detection routine
 *          that this rule keyword is supposed to encompass.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: If the detection test fails, this function *must* return a zero!
 *          On success, it calls the next function in the detection list 
 *
 ****************************************************************************/
int WifiFrameControlDetectorFunction(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
  
  WifiFrameControlData *plugin_data;
  plugin_data = otn->ds_list[PLUGIN_WIFI_FRAME_CONTROL_CHECK];

  if(!p->wifih){
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Not a WIFI frame\n"););
    return 0;
  }

  /* your detection function tests go here */
  if((plugin_data->not_flag && (p->wifih->frame_control != plugin_data->frame_control)) ||
	  (!plugin_data->not_flag && (p->wifih->frame_control == plugin_data->frame_control)))
    {
      return fp_list->next->OptTestFunc(p, otn, fp_list->next);
    }
#ifdef DEBUG
    else
    {
        /* you can put debug comments here or not */
        DebugMessage(DEBUG_PLUGIN,"No match\n");
    }
#endif

    /* if the test isn't successful, this function *must* return 0 */
    return 0;
}
