/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: sp_wifi_ssid.c,v 1.3 2003/07/27 07:03:04 andrew Exp $ */

/* sp_wifi_ssid
 * 
 * Purpose:
 *
 * Tests 802.11 frames SSID.
 *
 * Arguments:
 *   
 * {{ '!' | {} } <MAC> }
 *
 * Effect:
 *
 * Success on SSID match, failure otherwise.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "plugbase.h"
#include "parser.h"
#include "debug.h"
#include "util.h"
#include "plugin_enum.h"

#include "wifi_datatypes.h"

/* don't forget to include the name of this file in plugbase.h! */

extern char *file_name;  /* this is the file name from rules.c, generally used
                            for error messages */

extern int file_line;    /* this is the file line number from rules.c that is
                            used to indicate file lines for error messages */

typedef struct _WifiSsidData
{
  char *ssid;
  u_int8_t ssid_len;
  u_int8_t not_flag;
} WifiSsidData;

void WifiSsidInit(char *, OptTreeNode *, int);
void WifiSsidRuleParseFunction(char *, OptTreeNode *);
int WifiSsidDetectorFunction(Packet *, struct _OptTreeNode *, OptFpList *);

/****************************************************************************
 * 
 * Function: SetupWifiSsid()
 *
 * Purpose: Generic detection engine plugin template.  Registers the
 *          configuration function and links it to a rule keyword.  This is
 *          the function that gets called from InitPlugins in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void SetupWifiSsid()
{
    /* map the keyword to an initialization/processing function */
    RegisterPlugin("ssid", WifiSsidInit);

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Plugin: WifiSsid Setup\n"););
}


/****************************************************************************
 * 
 * Function: WifiSsidInit(char *, OptTreeNode *)
 *
 * Purpose: Generic rule configuration function.  Handles parsing the rule 
 *          information and attaching the associated detection function to
 *          the OTN.
 *
 * Arguments: data => rule arguments/data
 *            otn => pointer to the current rule option list node
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiSsidInit(char *data, OptTreeNode *otn, int protocol)
{
    /* allocate the data structure and attach it to the
       rule's data struct list */
    otn->ds_list[PLUGIN_WIFI_SSID_CHECK] = (WifiSsidData *) calloc(sizeof(WifiSsidData), sizeof(char));

    /* be sure to check that the protocol that is passed in matches the
       transport layer protocol that you're using for this rule! */

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    WifiSsidRuleParseFunction(data, otn);

    /* finally, attach the option's detection function to the rule's 
       detect function pointer list */
    AddOptFuncToList(WifiSsidDetectorFunction, otn);
}



/****************************************************************************
 * 
 * Function: WifiSsidRuleParseFunction(char *, OptTreeNode *)
 *
 * Purpose: This is the function that is used to process the option keyword's
 *          arguments and attach them to the rule's data structures.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiSsidRuleParseFunction(char *data, OptTreeNode *otn)
{
    WifiSsidData *ds_ptr;  /* data struct pointer */
	 
    /* set the ds pointer to make it easier to reference the option's
       particular data struct */
    ds_ptr = otn->ds_list[PLUGIN_WIFI_SSID_CHECK];

    /* manipulate the option arguments here */
    while(isspace((int) *data)) data++;
    
    if(*data == '!'){
      ds_ptr->not_flag = 1;
      data++;
    }
	 
    /* set the final option arguments here */
    
	 ds_ptr->ssid = data;
	 if((ds_ptr->ssid_len = strlen(data)) > 32)
		FatalError("%s (%d) => SSID exceeds maximum length \"%s\"\n", file_name, file_line, data);
}


/****************************************************************************
 * 
 * Function: WifiSsidDetectorFunction(char *, OptTreeNode *)
 *
 * Purpose: Use this function to perform the particular detection routine
 *          that this rule keyword is supposed to encompass.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: If the detection test fails, this function *must* return a zero!
 *          On success, it calls the next function in the detection list 
 *
 ****************************************************************************/
int WifiSsidDetectorFunction(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
  
  WifiSsidData *plugin_data;
  WifiElementID *ssid_id;
  u_int8_t *ssid;

  plugin_data = otn->ds_list[PLUGIN_WIFI_SSID_CHECK];

  if(!p->wifih){
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Not a WIFI frame\n"););
    return 0;
  }

  
  if(p->wifih->type != TYPE_MANAGEMENT){
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Frame is not TYPE_MANAGEMENT\n"););
	 return 0;
  }
  
  ssid_id = (u_int8_t *)p->wifih + _OFFSET_MANAGEMENT_BODY;
  ssid = (u_int8_t *)p->wifih + _OFFSET_MANAGEMENT_BODY;
  
  switch(p->wifih->stype){
  case STYPE_BEACON:
	 ssid_id +=_OFFSET_BEACON_SSID_ID;
	 ssid +=_OFFSET_BEACON_SSID;
	 break;
  case STYPE_ASSOCREQ:
	 ssid_id += _OFFSET_ASSOCREQ_SSID_ID;
	 ssid += _OFFSET_ASSOCREQ_SSID;
	 break;
	 
  case STYPE_REASSOC_REQ:
	 ssid_id += _OFFSET_REASSOC_REQ_SSID_ID;
	 ssid += _OFFSET_REASSOC_REQ_SSID;
	 break;

  case STYPE_PROBEREQ:
	 ssid_id += _OFFSET_PROBEREQ_SSID_ID;
	 ssid += _OFFSET_PROBEREQ_SSID;
	 break;
	 
  case STYPE_PROBERESP:
	 ssid_id += _OFFSET_PROBERESP_SSID_ID;
	 ssid += _OFFSET_PROBERESP_SSID;
	 break;
	 
  case STYPE_ASSOCRESP:
  case STYPE_REASSOC_RESP:
  case STYPE_ATIM:
  case STYPE_DISASSOC:
  case STYPE_AUTH:
  case STYPE_DEAUTH:
  default:
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Management frame subtype does not contain SSID\n"););
	 return 0;
  }

  /* your detection function tests go here */
  
  if((plugin_data->not_flag && (memcmp(ssid, plugin_data->ssid, plugin_data->ssid_len) != 0)) || 
	  (!plugin_data->not_flag && (memcmp(ssid, plugin_data->ssid, plugin_data->ssid_len) == 0)))
	 {
		return fp_list->next->OptTestFunc(p, otn, fp_list->next);
	 }
#ifdef DEBUG
  else
	 {
		/* you can put debug comments here or not */
		DebugMessage(DEBUG_PLUGIN,"No match\n");
	 }
#endif

  /* if the test isn't successful, this function *must* return 0 */
  return 0;
}



