/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/* $Id: sp_wifi_stype.c,v 1.5 2003/08/15 02:17:48 andrew Exp $ */

/* sp_wifi_stype
 * 
 * Purpose:
 *
 * Tests 802.11 frames for their frame subtype and implicitly tests the frame's type too
 * so that using a symbollic name for the subtype works in the assumed manner.
 *
 * Arguments:
 *   
 * {{ '!' | {} } { wifi_stype_t }
 *
 * Effect:
 *
 * Success on frame sub-type match, failure otherwise.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "plugbase.h"
#include "parser.h"
#include "debug.h"
#include "util.h"
#include "plugin_enum.h"


/* don't forget to include the name of this file in plugbase.h! */

extern char *file_name;  /* this is the file name from rules.c, generally used
                            for error messages */

extern int file_line;    /* this is the file line number from rules.c that is
                            used to indicate file lines for error messages */
#define NUM_WIFI_STYPES 40
char *_wifi_stype_strs[] = {
  "STYPE_ASSOCREQ",				  /* 0x00 */
  "STYPE_ASSOCRESP",				  /* 0x01 */
  "STYPE_REASSOC_REQ",			  /* 0x02 */
  "STYPE_REASSOC_RESP",			  /* 0x03 */
  "STYPE_PROBEREQ",				  /* 0x04 */
  "STYPE_PROBERESP",				  /* 0x05 */
  "STYPE_INVALID",				  /* 0x06 */
  "STYPE_INVALID",				  /* 0x07 */
  "STYPE_BEACON",					  /* 0x08 */
  "STYPE_ATIM",					  /* 0x09 */
  "STYPE_DISASSOC",				  /* 0x0a */
  "STYPE_AUTH",					  /* 0x0b */
  "STYPE_DEAUTH",					  /* 0x0c */
  "STYPE_INVALID",				  /* 0x0d */
  "STYPE_INVALID",				  /* 0x0e */
  "STYPE_INVALID",				  /* 0x0f */
  "STYPE_INVALID",				  /* 0x10 */
  "STYPE_INVALID",				  /* 0x11 */
  "STYPE_INVALID",				  /* 0x12 */
  "STYPE_INVALID",				  /* 0x13 */
  "STYPE_INVALID",				  /* 0x14 */
  "STYPE_INVALID",				  /* 0x15 */
  "STYPE_INVALID",				  /* 0x16 */
  "STYPE_INVALID",				  /* 0x17 */
  "STYPE_INVALID",				  /* 0x18 */
  "STYPE_INVALID",				  /* 0x19 */
  "STYPE_PSPOLL",					  /* 0x1a */
  "STYPE_RTS",						  /* 0x1b */
  "STYPE_CTS",						  /* 0x1c */
  "STYPE_ACK",						  /* 0x1d */
  "STYPE_CFEND",  				  /* 0x1e */
  "STYPE_CFEND_CFACK",			  /* 0x1f */
  "STYPE_DATA",					  /* 0x20 */
  "STYPE_CFACK",					  /* 0x21 */
  "STYPE_CFPOLL",					  /* 0x22 */
  "STYPE_CFACK_CFPOLL",			  /* 0x23 */
  "STYPE_NULLFUNC",				  /* 0x24 */
  "STYPE_CFACK_NULLFUNC",		  /* 0x25 */
  "STYPE_CFPOLL_NULLFUNC",		  /* 0x26 */
  "STYPE_CFACK_CFPOLL_NULLFUNC" /* 0x27 */
};

typedef enum {                  /* 802.11 management frame subtypes */
  STYPE_INVALID = -1,
  STYPE_ASSOCREQ = 0x00,		  /* Subtypes are already AND'ed with their corresponding Types */
  STYPE_ASSOCRESP = 0x01,		  /* to prevent false positives and the need for checking the type */
  STYPE_REASSOC_REQ = 0x02,	  /* independently */
  STYPE_REASSOC_RESP = 0x03,
  STYPE_PROBEREQ = 0x04,
  STYPE_PROBERESP = 0x05,
  STYPE_BEACON = 0x08,
  STYPE_ATIM = 0x09,
  STYPE_DIASSOC = 0x0a,
  STYPE_AUTH = 0x0b,
  STYPE_DEAUTH = 0x0c,

/* 802.11 control frame subtypes */
  STYPE_PSPOLL = 0x1a,
  STYPE_RTS = 0x1b,
  STYPE_CTS =  0x1c,
  STYPE_ACK = 0x1d,
  STYPE_CFEND = 0x1e,
  STYPE_CFEND_CFACK = 0x1f,

/* 802.11 data frame subtypes */
  STYPE_DATA = 0x20,
  STYPE_CFACK = 0x21,
  STYPE_CFPOLL = 0x22,
  STYPE_CFACK_CFPOLL = 0x23,
  STYPE_NULLFUNC = 0x24,
  STYPE_CFACK_NULLFUNC = 0x25,
  STYPE_CFPOLL_NULLFUNC = 0x26,
  STYPE_CFACK_CFPOLL_NULLFUNC = 0x27
} wifi_stype_t;

typedef struct _WifiStypeData
{
  wifi_stype_t stype;
  u_int8_t not_flag;

} WifiStypeData;

void WifiStypeInit(char *, OptTreeNode *, int);
void WifiStypeRuleParseFunction(char *, OptTreeNode *);
int WifiStypeDetectorFunction(Packet *, struct _OptTreeNode *, OptFpList *);

/****************************************************************************
 * 
 * Function: SetupWifiStype()
 *
 * Purpose: Generic detection engine plugin template.  Registers the
 *          configuration function and links it to a rule keyword.  This is
 *          the function that gets called from InitPlugins in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void SetupWifiStype()
{
    /* map the keyword to an initialization/processing function */
    RegisterPlugin("stype", WifiStypeInit);

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Plugin: WifiStype Setup\n"););
}


/****************************************************************************
 * 
 * Function: WifiStypeInit(char *, OptTreeNode *)
 *
 * Purpose: Generic rule configuration function.  Handles parsing the rule 
 *          information and attaching the associated detection function to
 *          the OTN.
 *
 * Arguments: data => rule arguments/data
 *            otn => pointer to the current rule option list node
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiStypeInit(char *data, OptTreeNode *otn, int protocol)
{
    /* allocate the data structure and attach it to the
       rule's data struct list */
  otn->ds_list[PLUGIN_WIFI_STYPE_CHECK] = (WifiStypeData *) calloc(sizeof(WifiStypeData), sizeof(char));

    /* be sure to check that the protocol that is passed in matches the
       transport layer protocol that you're using for this rule! */

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    WifiStypeRuleParseFunction(data, otn);

    /* finally, attach the option's detection function to the rule's 
       detect function pointer list */
    AddOptFuncToList(WifiStypeDetectorFunction, otn);
}



/****************************************************************************
 * 
 * Function: WifiStypeRuleParseFunction(char *, OptTreeNode *)
 *
 * Purpose: This is the function that is used to process the option keyword's
 *          arguments and attach them to the rule's data structures.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiStypeRuleParseFunction(char *data, OptTreeNode *otn)
{
  int i;
    WifiStypeData *ds_ptr;  /* data struct pointer */

    /* set the ds pointer to make it easier to reference the option's
       particular data struct */
    ds_ptr = otn->ds_list[PLUGIN_WIFI_STYPE_CHECK];

    /* manipulate the option arguments here */
	 while(isspace((int)*data)) data++;

	 if(*data == '!'){
		ds_ptr->not_flag = 1;
		data++;
	 }
	 
	 ds_ptr->stype = STYPE_INVALID;
	 
    /* set the final option arguments here */
	 for(i = 0; i < NUM_WIFI_STYPES; i++)
		if(strcmp(data, _wifi_stype_strs[i]) == 0 &&
			strcmp(data, "STYPE_INVALID") != 0){
		  ds_ptr->stype = i;
		  break;
		}
	 
	 if(ds_ptr->stype == STYPE_INVALID)
		FatalError("%s(%d) => Bad frame sub-type \"%s\"\n", file_name, file_line, data);
}


/****************************************************************************
 * 
 * Function: WifiStypeDetectorFunction(char *, OptTreeNode *)
 *
 * Purpose: Use this function to perform the particular detection routine
 *          that this rule keyword is supposed to encompass.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: If the detection test fails, this function *must* return a zero!
 *          On success, it calls the next function in the detection list 
 *
 ****************************************************************************/
int WifiStypeDetectorFunction(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
  WifiStypeData *wtd;
  u_int8_t stype;
  wtd = otn->ds_list[PLUGIN_WIFI_STYPE_CHECK];
  
  if(!p->wifih)
	 {
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Not WIFI\n"););
		return 0;
	 }
  
  stype = p->wifih->stype | (p->wifih->type << 4);
  
  if((wtd->not_flag && (stype != wtd->stype)) || 
	  (!wtd->not_flag && (stype == wtd->stype)))
	 {
		return fp_list->next->OptTestFunc(p, otn, fp_list->next);
    }
#ifdef DEBUG
    else
    {
		/* you can put debug comments here or not */
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"No match\n"););
    }
#endif

    /* if the test isn't successful, this function *must* return 0 */
    return 0;
}
