/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/



/* $Id: sp_wifi_type.c,v 1.2 2003/07/27 00:33:01 andrew Exp $ */
/* Snort Detection Plugin Source File WifiType */

/* sp_wifi_type
 * 
 * Purpose:
 *
 * Tests 802.11 frames for their frame type.
 *
 * Arguments:
 *   
 * {{ '!' | {} } { TYPE_MANAGEMENT | TYPE_CONTROL | TYPE_DATA } }
 *
 * Effect:
 *
 * Success on frame type match, failure otherwise.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>

#include "rules.h"
#include "decode.h"
#include "plugbase.h"
#include "parser.h"
#include "debug.h"
#include "util.h"
#include "plugin_enum.h"

#include "wifi_datatypes.h"

/* don't forget to include the name of this file in plugbase.h! */

extern char *file_name;  /* this is the file name from rules.c, generally used
                            for error messages */

extern int file_line;    /* this is the file line number from rules.c that is
                            used to indicate file lines for error messages */

char *_wifi_type_strs[] = {
  "TYPE_MANAGEMENT",
  "TYPE_CONTROL",
  "TYPE_DATA"
};

typedef struct _WifiTypeData
{
  wifi_type_t type;
  u_int8_t not_flag;

} WifiTypeData;

void WifiTypeInit(char *, OptTreeNode *, int);
void WifiTypeRuleParseFunction(char *, OptTreeNode *);
int WifiTypeDetectorFunction(Packet *, struct _OptTreeNode *, OptFpList *);

/****************************************************************************
 * 
 * Function: SetupWifiType()
 *
 * Purpose: Generic detection engine plugin template.  Registers the
 *          configuration function and links it to a rule keyword.  This is
 *          the function that gets called from InitPlugins in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void SetupWifiType()
{
    /* map the keyword to an initialization/processing function */
    RegisterPlugin("type", WifiTypeInit);

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Plugin: WifiType Setup\n"););
}


/****************************************************************************
 * 
 * Function: WifiTypeInit(char *, OptTreeNode *)
 *
 * Purpose: Generic rule configuration function.  Handles parsing the rule 
 *          information and attaching the associated detection function to
 *          the OTN.
 *
 * Arguments: data => rule arguments/data
 *            otn => pointer to the current rule option list node
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiTypeInit(char *data, OptTreeNode *otn, int protocol)
{
    /* allocate the data structure and attach it to the
       rule's data struct list */
  otn->ds_list[PLUGIN_WIFI_TYPE_CHECK] = (WifiTypeData *) calloc(sizeof(WifiTypeData), sizeof(char));

    /* be sure to check that the protocol that is passed in matches the
       transport layer protocol that you're using for this rule! */

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    WifiTypeRuleParseFunction(data, otn);

    /* finally, attach the option's detection function to the rule's 
       detect function pointer list */
    AddOptFuncToList(WifiTypeDetectorFunction, otn);
}



/****************************************************************************
 * 
 * Function: WifiTypeRuleParseFunction(char *, OptTreeNode *)
 *
 * Purpose: This is the function that is used to process the option keyword's
 *          arguments and attach them to the rule's data structures.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: void function
 *
 ****************************************************************************/
void WifiTypeRuleParseFunction(char *data, OptTreeNode *otn)
{
  int i;
    WifiTypeData *ds_ptr;  /* data struct pointer */

    /* set the ds pointer to make it easier to reference the option's
       particular data struct */
    ds_ptr = otn->ds_list[PLUGIN_WIFI_TYPE_CHECK];

    /* manipulate the option arguments here */
	 while(isspace((int)*data)) data++;

	 if(*data == '!'){
		ds_ptr->not_flag = 1;
		data++;
	 }
	 
	 ds_ptr->type = TYPE_INVALID;
	 
    /* set the final option arguments here */
	 for(i = 0; i < 3; i++)
		if(strcmp(data, _wifi_type_strs[i]) == 0){
		  ds_ptr->type = i;
		  break;
		}
	 
	 if(ds_ptr->type == TYPE_INVALID)
		FatalError("%s(%d) => Bad frame type \"%s\"\n", 
					  file_name, file_line, data);
}


/****************************************************************************
 * 
 * Function: WifiTypeDetectorFunction(char *, OptTreeNode *)
 *
 * Purpose: Use this function to perform the particular detection routine
 *          that this rule keyword is supposed to encompass.
 *
 * Arguments: data => argument data
 *            otn => pointer to the current rule's OTN
 *
 * Returns: If the detection test fails, this function *must* return a zero!
 *          On success, it calls the next function in the detection list 
 *
 ****************************************************************************/
int WifiTypeDetectorFunction(Packet *p, struct _OptTreeNode *otn, OptFpList *fp_list)
{
  WifiTypeData *wtd;
  wtd = otn->ds_list[PLUGIN_WIFI_TYPE_CHECK];
  
  if(!p->wifih)
	 {
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "Not WIFI\n"););
		return 0;
	 }
  
  if((wtd->not_flag && (p->wifih->type != wtd->type)) || 
	  (!wtd->not_flag && (p->wifih->type == wtd->type)))
	 {
		return fp_list->next->OptTestFunc(p, otn, fp_list->next);
    }
#ifdef DEBUG
    else
    {
		/* you can put debug comments here or not */
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"No match\n"););
    }
#endif

    /* if the test isn't successful, this function *must* return 0 */
    return 0;
}
