/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: spp_antistumbler.c,v 1.3 2005/06/25 05:46:21 andrew Exp $ */

/* spp_antistumbler
 * 
 * Purpose:
 *
 * Detects wireless stations actively scanning for APs
 *
 * Arguments:
 *
 * probe_reqs => number of probe requests during time delta it takes to trigger an alert
 *
 * probe_period => time period used to keep count of probe requests over 
 *
 * expire_timeout => number of seconds to wait for removing a MAC address from
 *                   the list of detected Netstumblers
 *
 * Effect:
 *
 * Sends an alert for each station it detects actively scanning for APs
 *
 * Comments:
 *
 * Any comments?
 *
 */


#include "generators.h"
#include "log.h"
#include "detect.h"
#include "decode.h"
#include "event.h"
#include "plugbase.h"
#include "parser.h"
#include "mstring.h"
#include "debug.h"
#include "util.h"


#ifndef __SPP_ANTISTUMBLER_C__
#define __SPP_ANTISTUMBLER_C__
#include "spp_antistumbler.h"
#endif


#define MODNAME "spp_antistumbler"

/* external globals from rules.c */
extern char *file_name;
extern int file_line;


AntiStumblerData antistumbler_data;

/*
 * Function: SetupAntiStumbler()
 *
 * Purpose: Registers the preprocessor keyword and initialization 
 *          function into the preprocessor list.  This is the function that
 *          gets called from InitPreprocessors() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void SetupAntiStumbler()
{
  /* link the preprocessor keyword to the init function in 
	  the preproc list */
  RegisterPreprocessor("antistumbler", AntiStumblerInit);
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: AntiStumbler is setup...\n");)
}


/*
 * Function: AntiStumblerInit(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void AntiStumblerInit(u_char *args)
{
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: AntiStumbler Initialized\n"););

    /* parse the argument list from the rules file */
    AntiStumblerParseArgs(args);

    /* Set the preprocessor function into the function list */
    AddFuncToPreprocList(AntiStumblerDetect);
/*     AddFuncToCleanExitList(PreprocCleanExitFunction); */
/*     AddFuncToRestartList(PreprocRestartFunction); */
}



/*
 * Function: AntiStumblerParseArgs(char *)
 *
 * Purpose: Process the preprocessor arguements from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */
void AntiStumblerParseArgs(char *args)
{
  int i, n, n_args, n_arg_args;
  char **arg_toks = NULL, **arg_args = NULL;
  
  /* init preproc data with defaults */
  antistumbler_data.probe_reqs = ANTISTUMBLER_PROBE_REQS;
  antistumbler_data.probe_period = ANTISTUMBLER_PROBE_PERIOD;
  antistumbler_data.expire_timeout = ANTISTUMBLER_EXPIRE_TIMEOUT;
  antistumbler_data.stumblers = NULL;
  
  if(args == NULL)
	 return;
  
  if(*args){
	 
	 arg_toks = mSplit(args, ",", 3, &n_args, 0);
	 if(n_args != 3){
		FatalError("%s: %s (%d) => antistumbler configuration format: probe_reqs <int>, probe_period <int>, expire_timeout <int>\n",
					  MODNAME, file_name, file_line);
	 }
	 
	 for(i = 0; i < n_args; i++){
		
		arg_args = mSplit(arg_toks[i], " ", 2, &n_arg_args, 0);
		if(n_arg_args != 2){
		  FatalError("%s: %s (%d) => antistumbler configuration format: probe_reqs <int>, probe_period <int>, expire_timeout <int>\n",
						 MODNAME, file_name, file_line);
		}

		if(strcmp(arg_args[0], "probe_reqs") == 0){
		  n = strtol(arg_args[1], NULL, 10);
		  if(n <= 0){
			 FatalError("%s: %s (%d) => \"probe_reqs\" must be greater than 0\n", MODNAME, file_name, file_line);
		  }
		  antistumbler_data.probe_reqs = n;
		}
		else if(strcmp(arg_args[0], "probe_period") == 0){
		  n = strtol(arg_args[1], NULL, 10);
		  if(n <= 0){
			 FatalError("%s: %s (%d) => \"probe_period\" must be greater than 0\n", MODNAME, file_name, file_line);
		  }
		  antistumbler_data.probe_period = n;
		}
		else if(strcmp(arg_args[0], "expire_timeout") == 0){
		  n = strtol(arg_args[1], NULL, 10);
		  if(n <= 0){
			 FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than 0\n", MODNAME, file_name, file_line);
		  }
		  antistumbler_data.expire_timeout = n;
		}  
		else{
		  FatalError("%s: %s (%d) => Invalid argument \"%s %s\"\n", MODNAME, file_name, file_line, arg_args[0], arg_args[1]);
		}
	 }
  }
  else{
	 FatalError("%s: %s (%d) => antistumbler configuration format: probe_reqs <int>, probe_period <int>, expire_timeout <int>\n",
					MODNAME, file_name, file_line);
  }
}

/*
 * Function: AntiStumblerDetect(Packet *)
 *
 * Purpose: Perform the preprocessor's intended function.  This can be
 *          simple (statistics collection) or complex (IP defragmentation)
 *          as you like.  Try not to destroy the performance of the whole
 *          system by trying to do too much....
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */
void AntiStumblerDetect(Packet *p)
{
  int on_list = 0;
  time_t time_curr;
  Stumbler *stumbler_ptr = NULL;
  Event event;
  char log_message[128];
  ProbeReq fixed_params;
  Beacon beacon_body;
  
  time_curr = time(NULL);
  
  if(p->wifih == NULL)
	 return;
  
  
  if(p->wifih->type == TYPE_MANAGEMENT && p->wifih->stype == STYPE_BEACON){
	 beacon_body.ssid_id = p->wifih + _OFFSET_BEACON_SSID_ID;
	 beacon_body.ssid = p->wifih + _OFFSET_BEACON_SSID;
	 
	 if(beacon_body.ssid_id->len == 15 && beacon_body.cap_info->cap_ibss){
		if(strncmp(beacon_body.ssid, "Wavelan Network", 15) == 0){
		  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: AdHoc network detected (SSID \"Wavelan Network\")\n"););
		  if(antistumbler_data.stumblers == NULL)
			 antistumbler_data.stumblers = AllocStumblerList();
		  
		  stumbler_ptr = InsertStumbler(p->wifih->addr3, antistumbler_data.stumblers);
		  
		  snprintf(log_message, 128, "Detected possible Kismet use from %.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
					  stumbler_ptr->addr[0], stumbler_ptr->addr[1], stumbler_ptr->addr[2], 
					  stumbler_ptr->addr[3], stumbler_ptr->addr[4], stumbler_ptr->addr[5]);
		  SetEvent(&event, GENERATOR_SPP_ANTISTUMBLER, STUMBLER_DETECTED, 1, 4, 2, 0);
/* 		  if(p->iph == NULL)	 */	  /* ugh, the horrible hack arises once again */
/* 			 p->iph = (IPHdr *)calloc(1, sizeof(IPHdr)); */
		  
		  CallAlertFuncs(p, log_message, NULL, &event);	/* send alert */
		}
	 }
	 return;
  }
  

  if(p->wifih->type != TYPE_MANAGEMENT || p->wifih->stype != STYPE_PROBEREQ)
	 return;
  
  fixed_params.ssid_id = p->pkt + _OFFSET_PROBEREQ_SSID_ID;
  fixed_params.ssid = p->pkt + _OFFSET_PROBEREQ_SSID;
/*   fixed_params.rates_id = fixed_params.ssid + fixed_params.ssid_id->len; */
/*   fixed_params.rates = fixed_params.rates_id + sizeof(ElementID); */
  
  if(fixed_params.ssid_id->len != 0){
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: SSID in probe request not NULL\n", MODNAME););
	 return;
  }
  
  if(antistumbler_data.stumblers == NULL)
	 antistumbler_data.stumblers = AllocStumblerList();
  
  stumbler_ptr = antistumbler_data.stumblers->head;
  while(stumbler_ptr){
	 
	 
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: STA:  %.2X:%.2X:%.2X:%.2X:%.2X:%.2X (probe_cnt: %d, ts_init: %d, ts_last: %d, ts_diff: %d)\n",
									 MODNAME, stumbler_ptr->addr[0], stumbler_ptr->addr[1], stumbler_ptr->addr[2], 
									 stumbler_ptr->addr[3], stumbler_ptr->addr[4], stumbler_ptr->addr[5],
									 stumbler_ptr->probe_cnt, stumbler_ptr->ts_init, stumbler_ptr->ts_last, 
									 stumbler_ptr->ts_last - stumbler_ptr->ts_init););
	 
	 if(memcmp(p->wifih->addr2, stumbler_ptr->addr, 6) == 0){ /* check to see if STA is on list */
		on_list = 1;
		stumbler_ptr->probe_cnt++;
		stumbler_ptr->ts_last = time_curr;
		
		
		if(stumbler_ptr->probe_cnt >= antistumbler_data.probe_reqs){
		  
		  if(stumbler_ptr->alert_flag == 0){
			 stumbler_ptr->alert_flag = 1;
			 
			 snprintf(log_message, 128, "Detected Netstumbler traffic from %.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
						 stumbler_ptr->addr[0], stumbler_ptr->addr[1], stumbler_ptr->addr[2], 
						 stumbler_ptr->addr[3], stumbler_ptr->addr[4], stumbler_ptr->addr[5]);
			 
			 SetEvent(&event, GENERATOR_SPP_ANTISTUMBLER, STUMBLER_DETECTED, 1, 4, 2, 0);
/* 			 if(p->iph == NULL) */		  /* horrible hack to get this to show up in ACID */
/* 				p->iph = (IPHdr *)calloc(1, sizeof(IPHdr)); */
			 
			 CallAlertFuncs(p, log_message, NULL, &event);	/* send alert */
		  }
		}
		
		if(stumbler_ptr->ts_last - stumbler_ptr->ts_init >= antistumbler_data.probe_period){
		  stumbler_ptr->ts_init = time_curr;
		  stumbler_ptr->probe_cnt = 1;
		}

	 }
	 if(time_curr  - stumbler_ptr->ts_last >= antistumbler_data.expire_timeout){
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Removing expired STA (%.2X:%.2X:%.2X:%.2X:%.2X:%.2X) from stumbler list\n",
										MODNAME, stumbler_ptr->addr[0], stumbler_ptr->addr[1], stumbler_ptr->addr[2], 
										stumbler_ptr->addr[3], stumbler_ptr->addr[4], stumbler_ptr->addr[5]););
		stumbler_ptr = RemoveStumbler(stumbler_ptr, antistumbler_data.stumblers);
		continue;
	 }
	 
	 stumbler_ptr = stumbler_ptr->next;
  }
  
  if(!on_list){
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding STA (%.2X:%.2X:%.2X:%.2X:%.2X:%.2X) to stumbler list\n", MODNAME,
									 p->wifih->addr2[0], p->wifih->addr2[1], p->wifih->addr2[2], 
									 p->wifih->addr2[3], p->wifih->addr2[4], p->wifih->addr2[5]););
	 
	 InsertStumbler(p->wifih->addr2, antistumbler_data.stumblers);
  }
}



  void PreprocCleanExitFunction(int signal)
{
       /* clean exit code goes here */
}

void PreprocRestartFunction(int signal)
{
       /* restart code goes here */
}


StumblerList *AllocStumblerList(void){
  return (StumblerList *)calloc(1, sizeof(StumblerList));
}


void DestroyStumblerList(StumblerList *list){
#warning "Implement DestroyStumblerList()"
}


/*
 * Function: InsertStumbler(u_int8_t *, StumblerList *)
 *
 * Purpose: Creates a Stumbler node from a byte array and inserts
 *          it into the specified StumblerList
 *
 * Arguments: addr => byte array containing a MAC address
 *            list => StumblerList to insert new node in
 *
 * Returns: pointer to newly inserted Stumbler node or NULL
 *          if bssid or list is NULL
 *
 */

Stumbler *InsertStumbler(u_int8_t *addr, StumblerList *list){
  Stumbler *stumbler_ptr = NULL;

  if(list == NULL || addr == NULL)
	 return NULL;
  
  /* allocate new Stumbler node */
  stumbler_ptr = (Stumbler *)malloc(sizeof(Stumbler));
  memcpy(stumbler_ptr->addr, addr, 6);
  stumbler_ptr->alert_flag = 0;
  stumbler_ptr->probe_cnt = 1;
  stumbler_ptr->ts_init = time(NULL);
  stumbler_ptr->ts_last = stumbler_ptr->ts_init;
  stumbler_ptr->next = NULL;
  stumbler_ptr->prev = NULL;
  
  if(list->head == NULL && list->tail == NULL){	/* if empty list, set head and tail */
	 list->head = stumbler_ptr;
	 list->tail = stumbler_ptr;
  }
  else{								  /* if already a tail, link new node in */
	 stumbler_ptr->prev = list->tail;
	 list->tail->next = stumbler_ptr;
	 list->tail = stumbler_ptr;
  }
  
  return stumbler_ptr;				  /* return pointer to new Stumbler node */
}



/*
 * Function: RemoveStumbler(Stumbler *, StumblerList *)
 *
 * Purpose: Removes a Stumbler node from a StumblerList
 *
 * Arguments: node => pointer to Stumbler node to be removed
 *            list => StumblerList to remove node from
 *
 * Returns: next Stumbler node in the list or NULL if one
 *          does not exist
 */

Stumbler *RemoveStumbler(Stumbler *node, StumblerList *list){
  Stumbler *stumbler_ptr;
  
  if(list == NULL)
	 return node;
  
  if(node == list->head){
	 if((list->head = node->next) == NULL)
		list->tail = NULL;
	 else
		node->next->prev = NULL;
  }
  else{
	 if((node->next->prev = node->next) == NULL)
		list->tail = node->prev;
	 else
		node->next->prev = node->prev;
  }
  
  stumbler_ptr = node->next;
  free(node);
  
  return stumbler_ptr;
}



