/*
** Copyright (C) 2004 Sebastien Gracia <wireless@sgdn.pm.gouv.fr>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** Copyright disclaimers:
** The Directeur Central de la scurit des systmes d'information hereby
** disclaims all copyright interest in the changes and enhancements
** made by sebastien Gracia.
** 5th Fev 2004.   
*/


/* spp_auth_flood
 * 
 * Purpose: Detects wireless access point flooded with auth frames.
 * 
 * auth_threshold => number of auth frames during time delta it takes
 *                  to trigger an alert
 *
 * expire_timeout => time period used to keep count of auth frames 
 * 
 * target_limit => maximum number of access point inserted inside authAP
 *                 mempool
 *  
 * prune_period => number of seconds to wait for looking after some decayed
 *                 flooded AP inside mempool
 *
 * Effect:
 *
 * Sends an alert for each access point having been detected as flooded
 * with auth frames.
 *
 * Comments:
 *
 * Any comments?
 *
 */

#include <errno.h>

#include "plugbase.h"
#include "mstring.h"
#include "debug.h"
#include "log.h"
#include "wifi_statetable.h"
#include "generators.h"
#include "ubi_BinTree.h"
#include "mempool.h"
#include "detect.h"
#include "util.h"


#ifndef __SPP_AUTH_FLOOD__C__
#define __SPP_AUTH_FLOOD_C__
#include "spp_auth_flood.h"
#endif

#define MODNAME "spp_auth_flood"

AuthFloodData auth_flood_data;
char logpath[STD_BUF];

/* external globals from rules.c */

extern char *file_name;
extern int file_line;


/*
 * Function: SetupAuthFlood()
 *
 * Purpose: Registers the preprocessor keyword and initialization 
 *          function into the preprocessor list.  This is the function that
 *          gets called from InitPreprocessors() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void SetupAuthFlood()
{
  /* link the preprocessor keyword to the init function in 
     the preproc list */

  RegisterPreprocessor("auth_flood", AuthFloodInit);
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: auth_flood is setup...\n"););
  return;
}


/*
 * Function: AuthFloodInit(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void AuthFloodInit(u_char *args)
{
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: auth_flood Initialized\n"););
  
  memset(&auth_flood_data, 0, sizeof(AuthFloodData));
  
  /* parse the argument list from the rules file */

  AuthFloodParseArgs(args);
  
  auth_flood_data.AuthAPsPtr = &auth_flood_data.AuthAPs;
  
  /* set up the flooded AP tree */

  ubi_trInitTree(auth_flood_data.AuthAPsPtr, authAPCompareFunc, 0);
  
    /* set up the node pools */

  if(mempool_init(&auth_flood_data.AuthAPPool,
		  auth_flood_data.target_limit,
		  sizeof(AuthAP))){
    FatalError("ERROR: Can't initialize mempool for flooded APs\n");
  }

  /* setup logpath */

  strncpy(logpath, pv.log_dir, STD_BUF);
  strncat(logpath, "/authflood.log", STD_BUF);  

  auth_flood_data.logfile = fopen(logpath, "a+");
  if(auth_flood_data.logfile == NULL){
    FatalError("Can't open logfile: %s", logpath);
  }
  auth_flood_data.offset_alert_flag = WSTInit(sizeof(u_int8_t));
  auth_flood_data.offset_nbr = WSTInit(sizeof(u_int16_t));
  auth_flood_data.offset_time = WSTInit(sizeof(struct timeval));
   
  /* Set the preprocessor function into the function list */

  AddFuncToPreprocList(AuthFloodDetect);

  AddFuncToCleanExitList(AuthFloodCleanExitFunction, &auth_flood_data);
  AddFuncToRestartList(AuthFloodRestartFunction, &auth_flood_data);
  return;
}


/*
 * Function: AuthFloodParseArgs(char *)
 *
 * Purpose: Process the preprocessor arguments from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */
void AuthFloodParseArgs(char *args)
{
  int i, n, n_args, n_arg_args;
  char **arg_toks = NULL, **arg_args = NULL;
  
  /* init preproc data with defaults */

  auth_flood_data.auth_threshold = AUTH_FLOOD_THRESHOLD;
  auth_flood_data.expire_timeout = AUTH_FLOOD_EXPIRE_TIMEOUT;
  auth_flood_data.target_limit = AUTH_FLOOD_TARGET_LIMIT;
  auth_flood_data.prune_period = AUTH_FLOOD_PRUNE_PERIOD;

  if(args == NULL){
    LogMessage("auth_flood initialized with defaults:\n");
    LogMessage("    %s: %u\n", AUTH_OPT_threshold, auth_flood_data.auth_threshold);
    LogMessage("    %s: %u\n", AUTH_OPT_EXPIRE_TIMEOUT, auth_flood_data.expire_timeout);
    LogMessage("    %s: %u\n", AUTH_OPT_TARGET_LIMIT, auth_flood_data.target_limit);
    LogMessage("    %s: %u\n", AUTH_OPT_PRUNE_PERIOD, auth_flood_data.prune_period);
  
    /* Update wifi_statetable timeout */
  
    WSTUpdateTimeout(auth_flood_data.expire_timeout);
    return;
  }
  
  arg_toks = mSplit(args, ",", 4, &n_args, 0);
  if(n_args != 4){
    FatalError("%s: %s (%d) => auth_flood configuration format:"
	       " auth_threshold <int>, expire_timeout <int>,\ntarget_limit <int>,"
	       " prune_period <int>\n",
	       MODNAME, file_name, file_line);
  }
     
  for(i = 0; i < n_args; i++){
    arg_args = mSplit(arg_toks[i], " ", 2, &n_arg_args, 0);
    if(n_arg_args != 2){
      FatalError("%s: %s (%d) => auth_flood configuration format:"
		 " auth_threshold <int>, expire_timeout <int>,\ntarget_limit <int>,"
		 " prune_period <int>\n",
		 MODNAME, file_name, file_line);
    }
    
    if(strcmp(arg_args[0], "auth_threshold") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"auth_threshold\" must be greater than 0\n"
		   , MODNAME, file_name, file_line);
      }
      auth_flood_data.auth_threshold = n;
    }
    else if(strcmp(arg_args[0], "expire_timeout") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      auth_flood_data.expire_timeout = n;
    }
    else if(strcmp(arg_args[0], "target_limit") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"target_limit\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      auth_flood_data.target_limit = n;
    }
    else if(strcmp(arg_args[0], "prune_period") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"prune_period\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      auth_flood_data.prune_period = n;
    }
    else{
      FatalError("%s: %s (%d) => Invalid argument \"%s %s\"\n",
		 MODNAME, file_name, file_line, arg_args[0], arg_args[1]);
    }    
  }
  
  if(auth_flood_data.expire_timeout <= auth_flood_data.prune_period){
    FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than prune_period\n",
	       MODNAME, file_name, file_line);
  }
    
  LogMessage("auth_flood arguments:\n");
  LogMessage("    %s: %u\n", AUTH_OPT_threshold, auth_flood_data.auth_threshold);
  LogMessage("    %s: %u\n", AUTH_OPT_EXPIRE_TIMEOUT, auth_flood_data.expire_timeout);
  LogMessage("    %s: %u\n", AUTH_OPT_TARGET_LIMIT, auth_flood_data.target_limit);
  LogMessage("    %s: %u\n", AUTH_OPT_PRUNE_PERIOD, auth_flood_data.prune_period);

  /* Update wifi_statetable timeout */

  WSTUpdateTimeout(auth_flood_data.expire_timeout);
  return;
}


/*
 * Function: AuthFloodDetect(Packet *)
 *
 * Purpose: Detect wireless auth flood traffic
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */
void AuthFloodDetect(Packet *p)
{
  wst_flag_type  flag_addr;
  u_int32_t pruned;
  u_int16_t *n_auth_ptr;
  u_int8_t *alert_flag_ptr;
  struct timeval wst_tv, tv_diff;
  Event event;
  char outstring[255];

  /* Cull any expired sessions out */
  
  if(p->pkth->ts.tv_sec >= (auth_flood_data.prune_time.tv_sec + auth_flood_data.prune_period)){
    DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			      "Pruning out AP flooded memory pool due to expired prune_period\n"););
      auth_flood_data.prune_time.tv_sec = p->pkth->ts.tv_sec;
      pruned=PruneAuthAPs(p->pkth->ts.tv_sec);
      DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			      "%u AP removed from AuthAP memory pool\n", pruned););
  }
  
  DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			  "AP flooded with auth frames count: %u\n",
			  ubi_trCount(auth_flood_data.AuthAPsPtr)););
  
  if(p->wifih == NULL){
    return;
  }

  /* Is it an auth frame? */

  if((p->wifih->type != TYPE_MANAGEMENT) || (p->wifih->stype != STYPE_AUTH)){
    return;
  }
    
  flag_addr = WST_FLAG_DESTADDR;
  
  /* Calls WSTMain to fill up p->wstidx.cache_addr[flag_addr] with memory address corresponding
   * with the place where destination MAC address (target) is recorded inside WST. */
  
  WSTMain(p, flag_addr);
  
  if( p->wstidx.cache_addr[flag_addr] == WST_FAILED){
    return; 
  }

  /* check wheter or not an alert is already reported */
  
  alert_flag_ptr = (u_int8_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
					       auth_flood_data.offset_alert_flag));

  /* n_auth_ptr points to the number of auth frame received by target 
   according to expire_timeout condition */

  n_auth_ptr = (u_int16_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
					      auth_flood_data.offset_nbr));
  if(*alert_flag_ptr){
    
    /* Alert has already been sent. Purpose is now to update WST time data in order
       to avoid concerned auth flooded AP to be removed from auth flooded AP
       memory pool throught PruneAuthAPs function and uptade received auth
       frames number */
    
    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + auth_flood_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    *n_auth_ptr = (*n_auth_ptr) + 1 ;
    return;
  }

  memcpy(&wst_tv, ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + auth_flood_data.offset_time)),
	 sizeof(struct timeval));
  tv_diff = WSTTimevalSubtract(p->pkth->ts, wst_tv);
  
  /* check wether or not dezs station should be considered as a new potential flooded AP.
   * Update WST data */

  if( tv_diff.tv_sec > auth_flood_data.expire_timeout ){
    
    /* dest station is a new potential AP flooded with auth frames.
     * Update WST data */
    
    *n_auth_ptr = 1; 
    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + auth_flood_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
  }
  else{
    *n_auth_ptr = (*n_auth_ptr) + 1 ;
  }
  
  /* check whether or not threshold is reached */
  
  if( *n_auth_ptr == auth_flood_data.auth_threshold){
    
    /* alert condition are reached  */

    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + auth_flood_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    (*alert_flag_ptr) = 1;
 
    snprintf(outstring, 255, AUTHFLOOD_PREFIX_STR
	     "\nAddr src: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x -> Addr dst: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x"
	     ", Bssid: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.",
	     p->wifih->addr2[0], p->wifih->addr2[1], p->wifih->addr2[2],
	     p->wifih->addr2[3], p->wifih->addr2[4], p->wifih->addr2[5],
	     p->wifih->addr1[0], p->wifih->addr1[1], p->wifih->addr1[2],
	     p->wifih->addr1[3], p->wifih->addr1[4], p->wifih->addr1[5],
	     p->wifih->addr3[0], p->wifih->addr3[1], p->wifih->addr3[2],
	     p->wifih->addr3[3], p->wifih->addr3[4], p->wifih->addr3[5]);
    
    DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD, "%s\n", outstring););
    
    SetEvent(&event, GENERATOR_SPP_AUTHFLOOD, AUTHFLOOD_DETECTED, 1, 0, 0, 0);
    CallAlertFuncs(p, outstring, NULL, &event);	/* send alert & log data */
    AuthAPLog(p, 0);
    
    /* insert dest station inside authAP mempool */
    
    AddAuthAP(p->wstidx.cache_addr[flag_addr]);
  }
  return;
}


/*
 * Function: int authAPCompareFunc(ubi_trItemPtrItemPtr,
 *                                 ubi_trNodePtr NodePtr)
 *
 * Purpose: Callback function used by splay trees to sort authAP node
 *
 * Arguments:
 *
 * Returns:
 *
 */
int authAPCompareFunc(ubi_trItemPtr ItemPtr, ubi_trNodePtr NodePtr)
{
  AuthAP *A = (AuthAP *) NodePtr;
  AuthAP *B = (AuthAP *) ItemPtr;
  int32_t memcmp_result;
  

  memcmp_result = memcmp(&A->addr[0], &B->addr[0], 6);
  if(memcmp_result < 0){
    return 1;
  }
  else if(memcmp_result > 0){
    return -1;
  }
  return 0;
}


/*
 * Function: void AddAuthAP(u_int32_t wst_addr)
 *
 * Purpose: Adds a new AP to the AP flooded with auth frames tree
 *
 * Arguments: wst_addr =>  memory address where MAC address is recorded inside wst.
 * 
 * Returns:
 *
 */

void AddAuthAP(u_int32_t wst_addr)
{
  AuthAP *authap = NULL;
  MemBucket *mb = NULL;

  /* borrow a AP node from the AP node pool */

  mb = mempool_alloc(&auth_flood_data.AuthAPPool);

  if(mb == NULL){
    DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD, "out of AP Nodes\n"););
    return;
  }
  
  authap = (AuthAP *) mb->data;
  authap->bucket = mb;

  /* fill in the authAP struct */
  authap->wst_addr = wst_addr;
  memcpy(&authap->addr[0], (u_int8_t *) wst_addr, 6);
  
  DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			  "Assigning a AP with MAC addr of"
			  " %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
			  authap->addr[0],authap->addr[1],authap->addr[2],authap->addr[3],authap->addr[4],authap->addr[5]););
   
  /* add this flooded AP to AP tree */

  if(ubi_sptInsert(auth_flood_data.AuthAPsPtr,
		   (ubi_btNodePtr)authap,
		   (ubi_btNodePtr)authap, NULL) == ubi_trFALSE){
    DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			    "Insert into flooded AP tree failed\n"););
  }
  else{
    DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			    "Insert into flooded AP tree successful\n"););
  }
  return;
}

  
/*
 * Function: u_int32_t PruneAuthAPs(u_int32_t now)
 *
 * Purpose: Checks whether or not AP inserted inside flooded AP tree are still
 * auth flooded or not.
 * If not, removes concerned APs from flooded AP tree and calls AUTHAPLogs function.
 *
 * Arguments: now => packet's capture time in sec
 *
 * Returns: number of APs removed from flooded AP tree
 *
 */
u_int32_t PruneAuthAPs(u_int32_t now)
{
  AuthAP *idx;
  u_int32_t pruned = 0;
  u_int32_t authap_sec;
  u_int8_t *alert_flag_ptr;

  DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			  "Prune flooded AP called now: "
			  " %u , count: %u\n",
			  now, ubi_trCount(auth_flood_data.AuthAPsPtr)););
  
  if(ubi_trCount(auth_flood_data.AuthAPsPtr) < 1){
    DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
			    "exit PruneAuthAP function since flooded AP tree"
			    " is free\n"););
    return(0);
  }    
  
  /* number of thing that need to be deleted */

  idx = (AuthAP *) ubi_btFirst((ubi_btNodePtr)auth_flood_data.AuthAPsPtr->root);
  
  if(idx == NULL){
    return(0);
  }
  
  do{
    authap_sec = *((u_int32_t *) (idx->wst_addr + auth_flood_data.offset_time));
    if((authap_sec + auth_flood_data.expire_timeout) < now){
      
      /* AP shouldn't be much more considered as a auth flooded one.
	 Removes it from authAP pool */
      
      alert_flag_ptr = (u_int8_t *) (idx->wst_addr + auth_flood_data.offset_alert_flag);
      *alert_flag_ptr = 0;
      
      AuthAPLog(NULL, idx->wst_addr);
      AuthAP *savidx = idx;
      if(ubi_trCount(auth_flood_data.AuthAPsPtr) > 1){
	idx = (AuthAP *) ubi_btNext((ubi_btNodePtr)idx);
	DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD,
				"pruning stale flooded AP\n"););
	DeleteAuthAP(savidx);
	pruned++;
      }
      else{
	DeleteAuthAP(savidx);
	pruned++;
	return(pruned);
      }
    }
    else{
      if(idx != NULL && ubi_trCount(auth_flood_data.AuthAPsPtr)){
	idx = (AuthAP *) ubi_btNext((ubi_btNodePtr)idx);
      }
      else{
	return pruned;
      }
    }
  } while(idx != NULL);
  
  return pruned;
}


/*
 * Function: void DeleteAuthAP(AuthAP *authap)
 *
 * Purpose: removed flooded auth AP from authAP mempool
 *
 * Arguments:
 *
 * Returns:
 *
 */
void DeleteAuthAP(AuthAP *authap)
{
AuthAP *oldauthap;

oldauthap = (AuthAP *) ubi_sptRemove(auth_flood_data.AuthAPsPtr,
  (ubi_btNodePtr) authap);
mempool_free(&auth_flood_data.AuthAPPool,authap->bucket);
  return;
}


/*
 * Function: void AUTHAPLog(Packet *p, u_int32_t wst_addr)
 *
 * Purpose: Called either when AP authentication flood is detected or finished.
 * Logs data to a text file: auth_flood_data.logfile.
 *
 * Arguments:
 *
 * Returns:
 *
 */
void AuthAPLog(Packet *p, u_int32_t wst_addr)
{
  char timestamp[TIMEBUF_SIZE];

  if(p){
  ts_print((struct timeval *) &p->pkth->ts, timestamp);
  fprintf(auth_flood_data.logfile, "\n%s Auth flood reported\n", timestamp);
  PrintWifiHeader(auth_flood_data.logfile, p);
  }
  else{
    ts_print((struct timeval *) (wst_addr + auth_flood_data.offset_time), timestamp);
    fprintf(auth_flood_data.logfile, "\n%s End Auth flood :"
	    " -> Addr dst: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x,\n"
	    "%u auth frame reported.\n",  
	    timestamp,
	    *((u_int8_t *)(wst_addr)), *((u_int8_t *)(wst_addr + 1)), *((u_int8_t *)(wst_addr + 2)),
	    *((u_int8_t *)(wst_addr + 3)), *((u_int8_t *)(wst_addr + 4)), *((u_int8_t *)(wst_addr + 5)),
	    *((u_int16_t*)(wst_addr + auth_flood_data.offset_nbr)));
  }
  fflush(auth_flood_data.logfile);
  return;
}


/*
 * Function: void AuthCleanExitFunction(int, void *)
 *
 * Purpose:
 * 
 * Arguments: 
 *
 * Returns:
 */
void AuthFloodCleanExitFunction(int signal, void *arg)
{
  AuthFloodData *dfd_ptr = (AuthFloodData *) arg;
  
  DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD, "AuthFloodCleanExitFunction\n"););

  if(fclose(dfd_ptr->logfile))
    LogMessage("%s: fclose function failed while snort exiting: %s\n",
	       MODNAME, strerror(errno));
  return;
}



/*
 * Function: void AuthFloodCleanRestartFunction(int, void *)
 *
 * Purpose:
 * 
 * Arguments: 
 *
 * Returns:
 */
void AuthFloodRestartFunction(int signal, void *arg)
{
 AuthFloodData *dfd_ptr = (AuthFloodData *) arg;
  
 DEBUG_WRAP(DebugMessage(DEBUG_AUTHFLOOD, "AuthFloodRestartFunction\n"););

  if(fclose(dfd_ptr->logfile))
    LogMessage("%s: fclose function failed while snort restarting: %s\n",
	       MODNAME, strerror(errno));
  return;
}







