/*
** Copyright (C) 2004 Sebastien Gracia <wireless@sgdn.pm.gouv.fr>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** Copyright disclaimers:
** The Directeur Central de la scurit des systmes d'information hereby
** disclaims all copyright interest in the changes and enhancements
** made by sebastien Gracia.
** 5th Fev 2004.
*/


/* spp_deauth_flood
 *
 * Purpose: Detects wireless station flooded with dehaut frames.
 *
 * deauth_threshold => number of deauth frames during time delta it takes
 *                    to trigger an alert
 *
 * expire_timeout => time period used to keep count of deauth frames 
 * 
 * target_limit => maximum number of station inserted inside deauthstation
 *                 mempool
 *  
 * prune_period => number of seconds to wait for looking after some decayed
 *                 deauthstations  inside mempool
 *
 * Effect:
 *
 * Sends an alert for each station having been detected as flooded
 * with deauth frames.
 *
 * Comments:
 *
 * Any comments?
 *
 */

#include <errno.h>

#include "plugbase.h"
#include "mstring.h"
#include "debug.h"
#include "log.h"
#include "wifi_statetable.h"
#include "generators.h"
#include "ubi_BinTree.h"
#include "mempool.h"
#include "detect.h"
#include "util.h"


#ifndef __SPP_DEAUTH_FLOOD__C__
#define __SPP_DEAUTH_FLOOD_C__
#include "spp_deauth_flood.h"
#endif

#define MODNAME "spp_deauth_flood"

DeauthFloodData deauth_flood_data;
char logpath[STD_BUF];

/* external globals from rules.c */

extern char *file_name;
extern int file_line;


/*
 * Function: SetupDeauthFlood()
 *
 * Purpose: Registers the preprocessor keyword and initialization 
 *          function into the preprocessor list.  This is the function that
 *          gets called from InitPreprocessors() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void SetupDeauthFlood()
{
  /* link the preprocessor keyword to the init function in 
     the preproc list */

  RegisterPreprocessor("deauth_flood", DeauthFloodInit);
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: deauth_flood is setup...\n"););
  return;
}


/*
 * Function: DeauthFloodInit(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void DeauthFloodInit(u_char *args)
{
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: deauth_flood Initialized\n"););
  
  memset(&deauth_flood_data, 0, sizeof(DeauthFloodData));
  
  /* parse the argument list from the rules file */

  DeauthFloodParseArgs(args);
  
  deauth_flood_data.DeauthstationsPtr = &deauth_flood_data.Deauthstations;
  
  /* set up the deauthenticated station tree */

  ubi_trInitTree(deauth_flood_data.DeauthstationsPtr, deauthstationCompareFunc, 0);
  
    /* set up the node pools */

  if(mempool_init(&deauth_flood_data.DeauthstationPool,
		  deauth_flood_data.target_limit,
		  sizeof(Deauthstation))){
    FatalError("ERROR: Can't initialize mempool for deauthenticated stations\n");
  }

  /* setup logpath */

  strncpy(logpath, pv.log_dir, STD_BUF);
  strncat(logpath, "/deauthflood.log", STD_BUF);  

  deauth_flood_data.logfile = fopen(logpath, "a+");
  if(deauth_flood_data.logfile == NULL){
    FatalError("Can't open logfile: %s", logpath);
  }

  deauth_flood_data.offset_alert_flag = WSTInit(sizeof(u_int8_t));
  deauth_flood_data.offset_nbr = WSTInit(sizeof(u_int16_t));
  deauth_flood_data.offset_time = WSTInit(sizeof(struct timeval));
   
  /* Set the preprocessor function into the function list */

  AddFuncToPreprocList(DeauthFloodDetect);

  AddFuncToCleanExitList(DeauthFloodCleanExitFunction, &deauth_flood_data);
  AddFuncToRestartList(DeauthFloodRestartFunction, &deauth_flood_data);
  return;
}


/*
 * Function: DeauthFloodParseArgs(char *)
 *
 * Purpose: Process the preprocessor arguments from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */
void DeauthFloodParseArgs(char *args)
{
  int i, n, n_args, n_arg_args;
  char **arg_toks = NULL, **arg_args = NULL;
  
  /* init preproc data with defaults */

  deauth_flood_data.deauth_threshold = DEAUTH_FLOOD_THRESHOLD;
  deauth_flood_data.expire_timeout = DEAUTH_FLOOD_EXPIRE_TIMEOUT;
  deauth_flood_data.target_limit = DEAUTH_FLOOD_TARGET_LIMIT;
  deauth_flood_data.prune_period = DEAUTH_FLOOD_PRUNE_PERIOD;

  if(args == NULL){
    LogMessage("deauth_flood initialized with defaults:\n");
    LogMessage("    %s: %u\n", DEAUTH_OPT_threshold, deauth_flood_data.deauth_threshold);
    LogMessage("    %s: %u\n", DEAUTH_OPT_EXPIRE_TIMEOUT, deauth_flood_data.expire_timeout);
    LogMessage("    %s: %u\n", DEAUTH_OPT_TARGET_LIMIT, deauth_flood_data.target_limit);
    LogMessage("    %s: %u\n", DEAUTH_OPT_PRUNE_PERIOD, deauth_flood_data.prune_period);

    /* Update wifi_statetable timeout */

    WSTUpdateTimeout(deauth_flood_data.expire_timeout);
    return;
  }

  arg_toks = mSplit(args, ",", 4, &n_args, 0);
  if(n_args != 4){
    FatalError("%s: %s (%d) => deauth_flood configuration format:"
	       " deauth_threshold <int>, expire_timeout <int>,\ntarget_limit <int>,"
	       " prune_period <int>\n",
	       MODNAME, file_name, file_line);
  }
  
  for(i = 0; i < n_args; i++){
    arg_args = mSplit(arg_toks[i], " ", 2, &n_arg_args, 0);
    if(n_arg_args != 2){
      FatalError("%s: %s (%d) => deauth_flood configuration format:"
		 " deauth_threshold <int>, expire_timeout <int>,\ntarget_limit <int>,"
		 " prune_period <int>\n",
		 MODNAME, file_name, file_line);
    }
    
    if(strcmp(arg_args[0], "deauth_threshold") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"deauth_threshold\" must be greater than 0\n"
		   , MODNAME, file_name, file_line);
      }
      deauth_flood_data.deauth_threshold = n;
    }
    else if(strcmp(arg_args[0], "expire_timeout") == 0){
	n = strtol(arg_args[1], NULL, 10);
	if(n <= 0){
	  FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than 0\n",
		     MODNAME, file_name, file_line);
	}
	deauth_flood_data.expire_timeout = n;
    }
    else if(strcmp(arg_args[0], "target_limit") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"target_limit\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      deauth_flood_data.target_limit = n;
    }
    else if(strcmp(arg_args[0], "prune_period") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"prune_period\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
	}
      deauth_flood_data.prune_period = n;
    }
    else{
      FatalError("%s: %s (%d) => Invalid argument \"%s %s\"\n",
		 MODNAME, file_name, file_line, arg_args[0], arg_args[1]);
      }
  }
  
  if(deauth_flood_data.expire_timeout <= deauth_flood_data.prune_period){
    FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than prune_period\n",
	       MODNAME, file_name, file_line);
  }

  LogMessage("deauth_flood arguments:\n");
  LogMessage("    %s: %u\n", DEAUTH_OPT_threshold, deauth_flood_data.deauth_threshold);
  LogMessage("    %s: %u\n", DEAUTH_OPT_EXPIRE_TIMEOUT, deauth_flood_data.expire_timeout);
  LogMessage("    %s: %u\n", DEAUTH_OPT_TARGET_LIMIT, deauth_flood_data.target_limit);
  LogMessage("    %s: %u\n", DEAUTH_OPT_PRUNE_PERIOD, deauth_flood_data.prune_period);

  /* Update wifi_statetable timeout */

  WSTUpdateTimeout(deauth_flood_data.expire_timeout);
  return;
}


/*
 * Function: DeauthFloodDetect(Packet *)
 *
 * Purpose: Detect wireless auth flood traffic
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */
void DeauthFloodDetect(Packet *p)
{
  wst_flag_type  flag_addr;
  u_int32_t pruned;
  u_int16_t *n_deauth_ptr;
  u_int8_t *alert_flag_ptr;
  struct timeval wst_tv, tv_diff;
  Event event;
  char outstring[255];

  /* Cull any expired sessions out */

  if(p->pkth->ts.tv_sec >= (deauth_flood_data.prune_time.tv_sec + deauth_flood_data.prune_period)){
    DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			      "Pruning out deauthenticated stations due to expired prune_period\n"););
    deauth_flood_data.prune_time.tv_sec = p->pkth->ts.tv_sec;
    pruned = PruneDeauthstations(p->pkth->ts.tv_sec);
    DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			    "%u stations removed from Deauthstation memory pool\n", pruned););
  }
  
  DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			  "Deauthenticated station count: %u\n",
			  ubi_trCount(deauth_flood_data.DeauthstationsPtr)););
  
  if(p->wifih == NULL){
    return;
  }
  
  /* Is it a deauth frame? */
  
  if((p->wifih->type != TYPE_MANAGEMENT) || (p->wifih->stype != STYPE_DEAUTH)){
    return;
  }

  flag_addr = WST_FLAG_DESTADDR;
  
  /* Calls WSTMain to fill up p->wstidx.cache_addr[flag_addr] with memory address corresponding
   * with the place where destination MAC address (target) is recorded inside WST. */
  
  WSTMain(p, flag_addr);
  
  if( p->wstidx.cache_addr[flag_addr] == WST_FAILED){
    return; 
  }
  
  /* check wheter or not an alert is already reported */
  
  alert_flag_ptr = (u_int8_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
					       deauth_flood_data.offset_alert_flag));
  
  /* n_deauth_ptr points to the number of deauth frame received by target 
     according to expire_timeout condition */
  
  n_deauth_ptr = (u_int16_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
					      deauth_flood_data.offset_nbr));
 
  if(*alert_flag_ptr){
    
    /* Alert has already been sent. Purpose is now to update WST time data in order
       to avoid concerned deauth station to be removed from deauth stations
       memory pool throught PruneDeauthstations function and uptade received deauth
       frames number */
    
    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + deauth_flood_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    *n_deauth_ptr = (*n_deauth_ptr) + 1 ;
    return;
  }
  
  memcpy(&wst_tv, ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + deauth_flood_data.offset_time)),
	 sizeof(struct timeval));
  tv_diff = WSTTimevalSubtract(p->pkth->ts, wst_tv);

  /* check wether or not station should be considered as a new potential deauthstation.
   * Update WST data */
  
  if( tv_diff.tv_sec > deauth_flood_data.expire_timeout ){

    /* station is a new potential deauthstation.
     * Update WST data */
    
    *n_deauth_ptr = 1; 
    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + deauth_flood_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
  }
  else {
    *n_deauth_ptr = (*n_deauth_ptr) + 1 ;
  }
  
  /* check whether or not threshold is reached */
  
  if( *n_deauth_ptr == deauth_flood_data.deauth_threshold){
    
    /* alert condition are reached  */

    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + deauth_flood_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    (*alert_flag_ptr) = 1;

    snprintf(outstring, 255, DEAUTHFLOOD_PREFIX_STR
	     "\nAddr src: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x -> Addr dst: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x"
	     ", Bssid: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.",
	     p->wifih->addr2[0], p->wifih->addr2[1], p->wifih->addr2[2],
	     p->wifih->addr2[3], p->wifih->addr2[4], p->wifih->addr2[5],
	     p->wifih->addr1[0], p->wifih->addr1[1], p->wifih->addr1[2],
	     p->wifih->addr1[3], p->wifih->addr1[4], p->wifih->addr1[5],
	     p->wifih->addr3[0], p->wifih->addr3[1], p->wifih->addr3[2],
	     p->wifih->addr3[3], p->wifih->addr3[4], p->wifih->addr3[5]);
    
    DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD, "%s\n", outstring););
    
    SetEvent(&event, GENERATOR_SPP_DEAUTHFLOOD, DEAUTHFLOOD_DETECTED, 1, 0, 0, 0);
    CallAlertFuncs(p, outstring, NULL, &event);	/* send alert & log data */
    DSLog(p, 0);
    
    /* insert station inside deauthstation mempool */
    
    AddDeauthStation(p->wstidx.cache_addr[flag_addr]);
  }
  return;
}


/*
 * Function: int deauthstationCompareFunc(ubi_trItemPtrItemPtr,
 *                                        ubi_trNodePtr NodePtr)
 *
 * Purpose: Callback function used by splay trees to sort deauthstation node
 *
 * Arguments:
 *
 * Returns:
 *
 */
int deauthstationCompareFunc(ubi_trItemPtr ItemPtr, ubi_trNodePtr NodePtr)
{
  Deauthstation *A = (Deauthstation *) NodePtr;
  Deauthstation *B = (Deauthstation *) ItemPtr;
  int32_t memcmp_result;
  

  memcmp_result = memcmp(&A->addr[0], &B->addr[0], 6);
  if(memcmp_result < 0){
    return 1;
  }
  else if(memcmp_result > 0){
    return -1;
  }
  return 0;
}

  
/*
 * Function: void AddDeauthStation(u_int32_t wst_addr)
 *
 * Purpose: Adds a new deauthenticated station to the deauthenticated station tree
 *
 * Arguments: wst_addr =>  memory address where MAC address is recorded inside wst.
 * 
 * Returns:
 *
 */
void AddDeauthStation(u_int32_t wst_addr)
{
  Deauthstation *ds = NULL;
  MemBucket *mb = NULL;

  /* borrow a deauthenticated station node from the deauthenticated station node pool */

  mb = mempool_alloc(&deauth_flood_data.DeauthstationPool);

  if(mb == NULL){
      DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD, "out of deauthenticated station Nodes\n"););
      return;
    }

  ds = (Deauthstation *) mb->data;
  ds->bucket = mb;

  /* fill in the deauthstation struct */

  ds->wst_addr = wst_addr;
  memcpy(&ds->addr[0], (u_int8_t *) wst_addr, 6);
  
  DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			  "Assigning a deauthenticated station with MAC addr of"
			  " %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
			  ds->addr[0],ds->addr[1],ds->addr[2],ds->addr[3],ds->addr[4],ds->addr[5]););
   
  /* add this deauthenticated station to deauthtenticated station tree */

  if(ubi_sptInsert(deauth_flood_data.DeauthstationsPtr,
		   (ubi_btNodePtr)ds,
		   (ubi_btNodePtr)ds, NULL) == ubi_trFALSE){
    DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			    "Insert into deauthenticated station tree failed\n"););
  }
  else{
    DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			    "Insert into deauthenticated station tree successful\n"););
  }
  return;
}

  
/*
 * Function: u_int32_t PruneDeauthstations(u_int32_t now)
 *
 * Purpose: Checks whether or not stations inserted inside deauthenticated
 * station tree are still deauthenticated or not.
 * If not, removes concerned stations from deauthenticated station tree and
 * calls DSLogs function.
 *
 * Arguments: now => packet's capture time in sec
 *
 * Returns: number of station removed from deauthenticated station tree
 *
 */
u_int32_t PruneDeauthstations(u_int32_t now)
{
  Deauthstation *idx;
  u_int32_t pruned = 0;
  u_int32_t ds_sec;
  u_int8_t *alert_flag_ptr;
  
  DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			  "Prune deauthenticated stations called now: "
			  " %u , count: %u\n",
			  now, ubi_trCount(deauth_flood_data.DeauthstationsPtr)););
 
  
  if(ubi_trCount(deauth_flood_data.DeauthstationsPtr) < 1){
    DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
			    "exit PruneDeauthstation function since deauthenticated"
			    " station tree is free\n"););
    return(0);
  }    
  
  /* number of thing that need to be deleted */

  idx = (Deauthstation *) ubi_btFirst((ubi_btNodePtr)deauth_flood_data.DeauthstationsPtr->root);
  
  if(idx == NULL){
    return(0);
  }
  
  do{
    ds_sec = *((u_int32_t *) (idx->wst_addr + deauth_flood_data.offset_time));
    if((ds_sec + deauth_flood_data.expire_timeout) < now){
      
      /* station shouldn't be much more considered as a deauthedtified one.
	 Removes it from deauthstation pool */
      
      alert_flag_ptr = (u_int8_t *) (idx->wst_addr + deauth_flood_data.offset_alert_flag);
      *alert_flag_ptr = 0;

      DSLog(NULL, idx->wst_addr);
      Deauthstation *savidx = idx;
      if(ubi_trCount(deauth_flood_data.DeauthstationsPtr) > 1){
	idx = (Deauthstation *) ubi_btNext((ubi_btNodePtr)idx);
	DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD,
				"pruning stale deauthenticated station\n"););
	DeleteDeauthstation(savidx);
	pruned++;
      }
      else{
	DeleteDeauthstation(savidx);
	pruned++;
	return(pruned);
      }
    }
    else{
      if(idx != NULL && ubi_trCount(deauth_flood_data.DeauthstationsPtr)){
	idx = (Deauthstation *) ubi_btNext((ubi_btNodePtr)idx);
      }
      else{
	return pruned;
      }
    }
  } while(idx != NULL);
  
  return pruned;
}


/*
 * Function: void DeleteDeauthstation(Deauthstation *ds)
 *
 * Purpose: removed ds from deauthstation mempool
 *
 * Arguments:
 *
 * Returns:
 *
 */
void DeleteDeauthstation(Deauthstation *ds)
{
  Deauthstation *oldds;
  
  oldds = (Deauthstation *) ubi_sptRemove(deauth_flood_data.DeauthstationsPtr,
					   (ubi_btNodePtr) ds);
  mempool_free(&deauth_flood_data.DeauthstationPool,ds->bucket);
  return;
}


/*
 * Function: void DSLog(Packet *p, u_int32_t wst_addr)
 *
 * Purpose: Called either when deauthentication flood is detected or finished.
 * Logs data to a text file: deauth_flood_data.logfile.
 *
 * Arguments:
 *
 * Returns:
 *
 */
void DSLog(Packet *p, u_int32_t wst_addr)
{
  char timestamp[TIMEBUF_SIZE];

  if(p){
  ts_print((struct timeval *) &p->pkth->ts, timestamp);
  fprintf(deauth_flood_data.logfile, "\n%s Deauth flood reported\n", timestamp);
  PrintWifiHeader(deauth_flood_data.logfile, p);
  }
  else{
    ts_print((struct timeval *) (wst_addr + deauth_flood_data.offset_time), timestamp);
    fprintf(deauth_flood_data.logfile, "\n%s End Deauth flood :"
	    " -> Addr dst: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x,\n"
	    "%u deauth frame reported.\n",  
	    timestamp,
	    *((u_int8_t *)(wst_addr)), *((u_int8_t *)(wst_addr + 1)), *((u_int8_t *)(wst_addr + 2)),
	    *((u_int8_t *)(wst_addr + 3)), *((u_int8_t *)(wst_addr + 4)), *((u_int8_t *)(wst_addr + 5)),
	    *((u_int16_t *)(wst_addr + deauth_flood_data.offset_nbr)));
  }
  fflush(deauth_flood_data.logfile);
  return;
}


/*
 * Function: void DeauthCleanExitFunction(int, void *)
 *
 * Purpose:
 * 
 * Arguments: 
 *
 * Returns:
 */
void DeauthFloodCleanExitFunction(int signal, void *arg)
{
  DeauthFloodData *dfd_ptr = (DeauthFloodData *) arg;
  
  DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD, "DeauthFloodCleanExitFunction\n"););

  if(fclose(dfd_ptr->logfile)){
    LogMessage("%s: fclose function failed while snort exiting: %s\n",
	       MODNAME, strerror(errno));
  }
  return;
}



/*
 * Function: void DeauthFloodCleanRestartFunction(int, void *)
 *
 * Purpose:
 * 
 * Arguments: 
 *
 * Returns:
 */
void DeauthFloodRestartFunction(int signal, void *arg)
{
 DeauthFloodData *dfd_ptr = (DeauthFloodData *) arg;
  
 DEBUG_WRAP(DebugMessage(DEBUG_DEAUTHFLOOD, "DeauthFloodRestartFunction\n"););

 if(fclose(dfd_ptr->logfile)){
    LogMessage("%s: fclose function failed while snort restarting: %s\n",
	       MODNAME, strerror(errno));
 }
  return;
}
