/*
** Copyright (C) 2004 Sebastien Gracia <wireless@sgdn.pm.gouv.fr>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** Copyright disclaimers:
** The Directeur Central de la scurit des systmes d'information hereby
** disclaims all copyright interest in the changes and enhancements
** made by sebastien Gracia.
** 5th Fev 2004.
*/


/* spp_macspoof
 * 
 * Purpose: Detects wireless MAC addresses involved in some MAC spoofed traffic.
 * 
 * <MACSPOOF_MASKED_ADDR> => list of MAC addresses excluded from wireless MAC spoofing
 *                           detection process
 *
 * tolerate_gap => tolerate missing frames betweeen two consecutive frames issued
 *                 from same MAC address
 *
 * threshold => number of abnormal sequence number gaps during time delta
 *              to trigger an alert
 *
 * expire_timeout => time period used to keep count of abnormal sequence number gap 
 * 
 * spoofed_addr_limit => maximum number of MAC addresses inserted inside MAC spoofed 
 *                       addresses mempool
 *  
 * prune_period => number of seconds to wait for looking after some decayed
 *                 MAC addresses inside mempool
 *
 * Effect:
 *
 * Sends an alert whenever MAC spoofing activity is detected.
 *
 * Comments:
 * - Because of some particuliar firmware or some electro_magnetic conditions, this
 * preprocessors won't fit to all wireless features inside network.
 * Those would generate too much false positive alert.
 * Therefore, I suggest macspoof preprocessos users to run snort a first time
 * in order to identify those particuliar features and thereafter to insert
 * them inside MaskedMACaddrlist.
 *
 * - Whenever an alert si raised, it doesn't mean anytime that concerned MAC address
 * is spoofed. Furthermore, It means that this MAC adress is involved in some MAC
 * spoofed traffic (either MAC spoofed or MAC spoofing address).
 * 
 * - whenever "end abnormal sequence number gap" is logged, it doesn't exactly mean that
 * MAC spoofing ended. For example, it would occured if a spoofer succeed to divert
 * a user from network while keeping using his MAC address. In such a case, neither
 * abnormal sequence number gap would be detected nor alert would be sended
 * while MAC spoofing still occur.
 *
 * Any comments?
 *
 */


#include <errno.h>

#include "plugbase.h"
#include "mstring.h"
#include "debug.h"
#include "log.h"
#include "wifi_statetable.h"
#include "generators.h"
#include "ubi_BinTree.h"
#include "mempool.h"
#include "detect.h"
#include "util.h"

#ifndef __SPP_MACSPOOF_C__
#define __SPP_MACSPOOF_C__
#include "spp_macspoof.h"
#endif

#define MODNAME "spp_macspoof"

MACSpoofData MACspoof_data;
char logpath[STD_BUF];

/* external globals from rules.c */

extern char *file_name;
extern int file_line;


/*
 * Function: SetupMACSpoof()
 *
 * Purpose: Registers the preprocessor keyword and initialization 
 *          function into the preprocessor list.  This is the function that
 *          gets called from InitPreprocessors() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void SetupMACSpoof()
{
  /* link the preprocessor keyword to the init function in 
	  the preproc list */

  RegisterPreprocessor("macspoof", MACSpoofInit);
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: macspoof is setup...\n"););
  return;
}


/*
 * Function: MACSpoofInit(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void MACSpoofInit(u_char *args)
{
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: macspoof Initialized\n"););
  
  memset(&MACspoof_data, 0, sizeof(MACSpoofData));
  
  /* parse the argument list from the rules file */
  
  MACSpoofParseArgs(args);
  
  MACspoof_data.spoofed_MACaddrsPtr = &MACspoof_data.spoofed_MACaddrs;
  
  /* set up the MAC spoofed address tree */
  
  ubi_trInitTree(MACspoof_data.spoofed_MACaddrsPtr, MACSpoofedAddrCompareFunc, 0);
  
  /* set up the node pools */
  
  if(mempool_init(&MACspoof_data.spoofed_MACaddrPool,
		  MACspoof_data.spoofed_addr_limit,
		  sizeof(Spoofed_MACaddr))){
    FatalError("ERROR: Can't initialize mempool for MACSpoofed addresses\n");
  }
  
  /* setup logpath */
  
  strncpy(logpath, pv.log_dir, STD_BUF);
  strncat(logpath, "/MACspoof.log", STD_BUF);  
  
  MACspoof_data.logfile = fopen(logpath, "a+");
  if(MACspoof_data.logfile == NULL){
    FatalError("Can't open logfile: %s", logpath);
  }

  MACspoof_data.offset_alert_flag = WSTInit(sizeof(u_int8_t));
  MACspoof_data.offset_seq_number = WSTInit(sizeof(u_int16_t));
  MACspoof_data.offset_gap_number = WSTInit(sizeof(u_int16_t));
  MACspoof_data.offset_time = WSTInit(sizeof(struct timeval));
  
  /* Set the preprocessor function into the function list */
  
  AddFuncToPreprocList(MACSpoofDetect);
  
  AddFuncToCleanExitList(MACSpoofCleanExitFunction, &MACspoof_data);
  AddFuncToRestartList(MACSpoofRestartFunction, &MACspoof_data);
  return;
}


/*
 * Function: MACSpoofParseArgs(char *)
 *
 * Purpose: Process the preprocessor arguments from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */
void MACSpoofParseArgs(char *args)
{
  int i, n, n_args, n_arg_args;
  char **arg_toks = NULL, **arg_args = NULL;
  MACADDR *MAC_addr_ptr;

  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			  "Parsing macspoof preprocessor arguments\n"););
    
  /* init preproc data with defaults */

  MACspoof_data.tolerate_gap = MACSPOOF_TOLERATE_GAP;
  MACspoof_data.threshold = MACSPOOF_THRESHOLD;
  MACspoof_data.expire_timeout = MACSPOOF_EXPIRE_TIMEOUT;
  MACspoof_data.spoofed_addr_limit = MACSPOOF_SPOOFED_ADDR_LIMIT;
  MACspoof_data.prune_period = MACSPOOF_PRUNE_PERIOD;

  if(args == NULL) {
    LogMessage("macspoof initialized with defaults:\n");
    LogMessage("    %s: none\n", MACSPOOF_OPT_MASKED_MAC_ADDR);
    LogMessage("    %s: %u\n", MACSPOOF_OPT_TOLERATE_GAP, MACspoof_data.tolerate_gap);
    LogMessage("    %s: %u\n", MACSPOOF_OPT_THRESHOLD, MACspoof_data.threshold);
    LogMessage("    %s: %u\n", MACSPOOF_OPT_EXPIRE_TIMEOUT, MACspoof_data.expire_timeout);
    LogMessage("    %s: %u\n", MACSPOOF_OPT_SPOOFED_ADDR_LIMIT, MACspoof_data.spoofed_addr_limit);
    LogMessage("    %s: %u\n", MACSPOOF_OPT_PRUNE_PERIOD, MACspoof_data.prune_period);

    /* Update wifi_statetable timeout */

    WSTUpdateTimeout(MACspoof_data.expire_timeout);
    return;
  }

  /* Build masked MAC address list */

  MACspoof_data.masked_MACaddrlist = BuildMaskedMACaddrlist(&args);  
  
  if(MACspoof_data.masked_MACaddrlist == NULL){
    args += 4;
  }
  else{
    MACspoof_data.MaskedMACaddrlist_flag = 1;
  }

  /* finally process tolerate_gap, threshold, expire_timeout,spoofed_addr_limit and
     prune_period arguments */
    
  arg_toks = mSplit(args, ",", 5, &n_args, 0);
  if(n_args != 5){
    FatalError("%s: %s (%d) => macspoof configuration format:"
	       " tolerate_gap <int>, threshold <int>, expire_timeout <int>,\n"
	       "spoofed_addr_limit <int>, prune_period <int>\n",
	       MODNAME, file_name, file_line);
  }
  
  for(i = 0; i < n_args; i++) {
    arg_args = mSplit(arg_toks[i], " ", 2, &n_arg_args, 0);
    if(n_arg_args != 2){
      FatalError("%s: %s (%d) => macspoof configuration format:"
		 " tolerate_gap <int>, threshold <int>, expire_timeout <int>,\n"
		 "spoofed_addr_limit <int>, prune_period <int>\n",
		 MODNAME, file_name, file_line);
    }
    
    if(strcmp(arg_args[0], "tolerate_gap") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"tolerate_gap\" must be greater than 0\n"
		   , MODNAME, file_name, file_line);
      }
      MACspoof_data.tolerate_gap = n;
    }
    else if(strcmp(arg_args[0], "threshold") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 1){
	FatalError("%s: %s (%d) => \"threshold\" must be greater than 1\n"
		   , MODNAME, file_name, file_line);
      }
      MACspoof_data.threshold = n;
    }
    else if(strcmp(arg_args[0], "expire_timeout") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      MACspoof_data.expire_timeout = n;
    }
    else if(strcmp(arg_args[0], "spoofed_addr_limit") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"spoofed_addr_limit\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      MACspoof_data.spoofed_addr_limit = n;
    }
    else if(strcmp(arg_args[0], "prune_period") == 0){
      n = strtol(arg_args[1], NULL, 10);
      if(n <= 0){
	FatalError("%s: %s (%d) => \"prune_period\" must be greater than 0\n",
		   MODNAME, file_name, file_line);
      }
      MACspoof_data.prune_period = n;
    }
    else{
      FatalError("%s: %s (%d) => Invalid argument \"%s %s\"\n",
		 MODNAME, file_name, file_line, arg_args[0], arg_args[1]);
    }
  }

  if(MACspoof_data.expire_timeout <= MACspoof_data.prune_period){
    FatalError("%s: %s (%d) => \"expire_timeout\" must be greater than prune_period\n",
	       MODNAME, file_name, file_line);
  }
    
  LogMessage("macspoof arguments:\n");
  if(MACspoof_data.MaskedMACaddrlist_flag){
    MAC_addr_ptr = MACspoof_data.masked_MACaddrlist->head;
    LogMessage("    %s: ", MACSPOOF_OPT_MASKED_MAC_ADDR);
    LogMessage("%.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
	       MAC_addr_ptr->MAC_addr[0], MAC_addr_ptr->MAC_addr[1], MAC_addr_ptr->MAC_addr[2],
	       MAC_addr_ptr->MAC_addr[3], MAC_addr_ptr->MAC_addr[4], MAC_addr_ptr->MAC_addr[5]);
    MAC_addr_ptr = MAC_addr_ptr->next; 
    while(MAC_addr_ptr){
      LogMessage("                     %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
		 MAC_addr_ptr->MAC_addr[0], MAC_addr_ptr->MAC_addr[1], MAC_addr_ptr->MAC_addr[2],
		   MAC_addr_ptr->MAC_addr[3], MAC_addr_ptr->MAC_addr[4], MAC_addr_ptr->MAC_addr[5]);
      MAC_addr_ptr = MAC_addr_ptr->next; 
    }
  }
  else{
      LogMessage("    %s: none\n", MACSPOOF_OPT_MASKED_MAC_ADDR);
  }
  LogMessage("    %s: %u\n", MACSPOOF_OPT_TOLERATE_GAP, MACspoof_data.tolerate_gap);
  LogMessage("    %s: %u\n", MACSPOOF_OPT_THRESHOLD, MACspoof_data.threshold);
  LogMessage("    %s: %u\n", MACSPOOF_OPT_EXPIRE_TIMEOUT, MACspoof_data.expire_timeout);
  LogMessage("    %s: %u\n", MACSPOOF_OPT_SPOOFED_ADDR_LIMIT, MACspoof_data.spoofed_addr_limit);
  LogMessage("    %s: %u\n", MACSPOOF_OPT_PRUNE_PERIOD, MACspoof_data.prune_period);

  /* Update wifi_statetable timeout */

  WSTUpdateTimeout(MACspoof_data.expire_timeout);
  return;
}

  
/*
 * Function: MACSpoofDetect(Packet *)
 *
 * Purpose: Detect wireless frames with spoofed source address excluding ones specified
 *          by the MACSPOOF_MASKEDC_ADDR variable in snort.conf
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */
void MACSpoofDetect(Packet *p)
{
  wst_flag_type  flag_addr;
  u_int32_t pruned;
  u_int16_t *gap_number_ptr, *seq_number_ptr, gap;
  u_int8_t *alert_flag_ptr;
  int8_t src_offset;
  struct timeval wst_tv, tv_diff;
  Event event;
  char outstring[255];
  MACADDR *MAC_addr_ptr;

  /* Cull any expired MAC spoofed addresses out */  
  
  if(p->pkth->ts.tv_sec >= (MACspoof_data.prune_time.tv_sec + MACspoof_data.prune_period)){
          DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			      "Pruning out MACSpoofed addr due to expired prune_period\n"););
      MACspoof_data.prune_time.tv_sec = p->pkth->ts.tv_sec;
      pruned=PruneMACSpoofedAddrs(p->pkth->ts.tv_sec);
      DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			      "%u addr removed from MACSpoofed addr memory pool\n", pruned););
    }
  
  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			  "MACSPoofed addr count: %u\n",
			  ubi_trCount(MACspoof_data.spoofed_MACaddrsPtr)););
  
  if(p->wifih == NULL){
    return;
  }
  
  flag_addr = WST_FLAG_SRCADDR;
  
  if(MACspoof_data.MaskedMACaddrlist_flag){
    src_offset = WSTDetermineMACAddrOffset(p, flag_addr);
    MAC_addr_ptr = MACspoof_data.masked_MACaddrlist->head;
    do{      
      if(!memcmp(MAC_addr_ptr->MAC_addr, (u_int8_t *) p->wifih + src_offset, 6)){
	DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
				"Packet ignored due to src MAC addr: %2x:%.2x:%.2x:%.2x:%.2x:%.2x "
				"included inside masked MAC addr list\n",
				*((u_int8_t *) p->wifih + src_offset),
				*((u_int8_t *) p->wifih + src_offset + 1),
				*((u_int8_t *) p->wifih + src_offset + 2),
				*((u_int8_t *) p->wifih + src_offset + 3),
				*((u_int8_t *) p->wifih + src_offset + 4),
				*((u_int8_t *) p->wifih + src_offset + 5)););
	return;
      }
      MAC_addr_ptr = MAC_addr_ptr->next;
    } while(MAC_addr_ptr);
  }
  
  /* Calls WSTMain to fill up p->wstidx.cache_addr[flag_addr] with memory address corresponding
   * with the place where source MAC address is recorded inside WST. */
  
  WSTMain(p, flag_addr);
  
  if( p->wstidx.cache_addr[flag_addr] == WST_FAILED) {
    return; 
  }
    
  /* seq_number_ptr points to the last wireless frame sequence number issued from
     source MAC address */
  
  seq_number_ptr = (u_int16_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
						MACspoof_data.offset_seq_number));
  
  /* gap represent the difference between packet's sequence number and corresponding sequence
     number recorded inside WST minus one. Furthermore, it should be considered as the total
     missing frames refered to 802.11 protocol.
     This means that two consecutive frames with same sequence number issued from same MAC addr
     will generate a gab equal to 4095. */

  gap = (4095 + p->wifih->seqnum - (*seq_number_ptr)) %4096;
  
  /* Check whether or not gap should be considered as normal refered to preprocessor's argument
     tolerate_gap */
  
  if( gap <= MACspoof_data.tolerate_gap){
    
    /* packet sequence number fill up 802.11 protocol according to preprocessor arguments */
    
    *seq_number_ptr = p->wifih->seqnum;
    return;
  }
  
  /* From now, Packet is considered as abnormal */
  
  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			  "Missing frame abnormal count issued from MAC addr: "
			  "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
			  "received frame seq nbr: %u, recorded seq nbr: %u\n\n",
			  *((u_int8_t *) p->wstidx.cache_addr[flag_addr]),
			  *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 1)),
			  *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 2)),
			  *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 3)),
			  *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 4)),
			  *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 5)),
			  p->wifih->seqnum, *seq_number_ptr););
  
  /* check wheter or not an alert is already reported */
  
  alert_flag_ptr = (u_int8_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
					       MACspoof_data.offset_alert_flag));
  
  if(*alert_flag_ptr){
    
    /* Alert has already been sent. Purpose is now to update WST time data in order
       to avoid concerned spoofed MAC address to be removed from MAC spoofed address
       memory pool throught PruneMACSpoofedAddrs function */
    
    *seq_number_ptr = p->wifih->seqnum;
    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + MACspoof_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    return;
  }
  
  /* gap_number_ptr points to the number of abnormal gap associated with source MAC addr
     inside WST */
  
  gap_number_ptr = (u_int16_t *) ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] +
						MACspoof_data.offset_gap_number));
  
  /* Check whether or not data recorded inside WST should be considered as decayed for
     this preprocessor */
  
  memcpy(&wst_tv, ((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + MACspoof_data.offset_time)),
	 sizeof(struct timeval));
  tv_diff = WSTTimevalSubtract(p->pkth->ts, wst_tv);
  
  if( tv_diff.tv_sec > MACspoof_data.expire_timeout ){
    
    /* data recorded inside WST are considered as decayed for this preprocessor.
     Uptade them. */
    
    *gap_number_ptr = 1; 
    *seq_number_ptr = p->wifih->seqnum;
    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + MACspoof_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    return;
  }
  
  (*gap_number_ptr)++;
  *seq_number_ptr = p->wifih->seqnum;
  
  /* checks if alert condition are reached  */

  if((*gap_number_ptr) == MACspoof_data.threshold){

    /* alert condition are reached  */

    memcpy((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + MACspoof_data.offset_time),
	   &p->pkth->ts, sizeof(struct timeval));
    (*alert_flag_ptr) = 1;
    
    snprintf(outstring, 255, MACSPOOF_PREFIX_STR "\nabnormal seq number gap "
	     "occurs on addr: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
	     *((u_int8_t *) p->wstidx.cache_addr[flag_addr]),
	     *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 1)),
	     *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 2)),
	     *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 3)),
	     *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 4)),
	     *((u_int8_t *) (p->wstidx.cache_addr[flag_addr] + 5)));
    DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF, "%s\n", outstring););
    
    /* send alert & log data */
    
    SetEvent(&event, GENERATOR_SPP_MACSPOOF, MACSPOOF_DETECTED, 1, 0, 0, 0);
    CallAlertFuncs(p, outstring, NULL, &event);
    MACSpoofLog(p, p->wstidx.cache_addr[flag_addr]);
    
    /* insert station inside deauthstation mempool */
    
    AddMACSpoofedAddr(p->wstidx.cache_addr[flag_addr]);
  }
  return;
}


/*
 * Function: MaskedMACaddrlist *BuildMaskedMACaddrlist(char **args_ptr)
 *
 * Purpose: Parses MAC address argument from the preprocessor's arguments string into a 
 *          MaskedMACaddrlist and points it to the end of the MAC address argument
 *
 * Arguments: args_ptr => pointer to a string containing the preprocessor's arguments
 *
 * Returns: a MaskedMACaddrlist of MAC addresses to exclude from alerts
 *
 */
MaskedMACaddrlist *BuildMaskedMACaddrlist(char **args_ptr)
{
  char *end_ptr;
  char *args = *args_ptr;
  MaskedMACaddrlist *masked_MACaddrlist = NULL;
  u_int8_t MAC_addr[6];
  
  while(*args && isspace(*args))
    args++;

  /* checks if MAC address argument is none which means that no MAC address
     will be exclude from alerts */

  if(memcmp(args, "none", 4) == 0){
    return NULL;
  }
   
  masked_MACaddrlist = AllocMaskedMACaddrlist(); 

  /* parse the MAC addr or list of MAC addrs */
  
  if(args[0] == '['){
    for(end_ptr = ++args; *end_ptr && *end_ptr != ']'; end_ptr++){
      if(*end_ptr == ','){
	*end_ptr = '\0';
	if(ParseMACAddr(MAC_addr, args) < 0){
	  FatalError("%s: %s (%d) => Error parsing MAC addr \"%s\"\n",
		     MODNAME, file_name, file_line, args);
	}
	DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to masked MAC list\n", 
				MODNAME, MAC_addr[0],  MAC_addr[1],  MAC_addr[2],
				MAC_addr[3],  MAC_addr[4],  MAC_addr[5]););

	InsertMaskedMACaddr(MAC_addr, masked_MACaddrlist);
	args = end_ptr + 1;
      }
    }

    *end_ptr = '\0';
    if(ParseMACAddr(MAC_addr, args) < 0){
      FatalError("%s: %s (%d) => error parsing masked MAC addresses \"%s\"\n",
		 MODNAME, file_name, file_line, args);
    }
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to masked MAC list\n", 
			    MODNAME, MAC_addr[0],  MAC_addr[1],  MAC_addr[2],
			    MAC_addr[3],  MAC_addr[4],  MAC_addr[5]););

    InsertMaskedMACaddr(MAC_addr, masked_MACaddrlist);
    args = end_ptr + 1;
    while(*args && *args != ','){
      args++;
    }
    if(*args){
      args++;
    }
  }
  else{
    for(end_ptr = args; *end_ptr && *end_ptr != ','; end_ptr++);
    
    *end_ptr ='\0';
    if(ParseMACAddr(MAC_addr, args) < 0){
      FatalError("%s: %s (%d) => error parsing masked MAC addresses \"%s\"\n",
		 MODNAME, file_name, file_line, args);
    }
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to masked MAC list\n", 
			    MODNAME, MAC_addr[0],  MAC_addr[1],  MAC_addr[2],
			    MAC_addr[3],  MAC_addr[4],  MAC_addr[5]););

    InsertMaskedMACaddr(MAC_addr, masked_MACaddrlist);
    args = end_ptr + 1;
  }
  
  *args_ptr = args;
  
  return masked_MACaddrlist;
}


/*
 * Function: MaskedMACaddrlist *AllocMaskedMACaddrlist(void)
 *
 * Purpose: Allocates a MaskedMACaddrlist
 *
 * Arguments: None
 *
 * Returns: A pointer to the newly allocated MaskedMACaddrlist or NULL or -1 if failed
 *
 */
MaskedMACaddrlist *AllocMaskedMACaddrlist(void)
{
  return (MaskedMACaddrlist *)calloc(1, sizeof(MaskedMACaddrlist));
}


int ParseMACAddr(u_int8_t *MACaddr, char *MAC_str)
{
  int i = 0;
  char *next = MAC_str;

  while(*next){
    MACaddr[i] = strtol(MAC_str, &next, 16);

    if(MAC_str != next && i < 6){
      MAC_str = next + 1;
      i++;
    }
    else{
      return -1;
    }
  }
  return 0;
}


/*
 * Function: MACADDR *InsertMaskedMACaddr(u_int8_t *MAC_addr, MaskedMACaddrlist *list)
 *
 * Purpose: Creates a MACADDR node from a byte array and inserts it into the
 *          specified MaskedMACaddrlist
 *
 * Arguments: MAC_addr => byte array containing a MAC address
 *            list => MaskedMACaddrlist to insert new node in
 *
 * Returns: A pointer to newly inserted MACADDR node or NULL if MAC_addr or 
 *          list is NULL
 */
MACADDR *InsertMaskedMACaddr(u_int8_t *MAC_addr, MaskedMACaddrlist *list)
{
  MACADDR *MACAddr_ptr = NULL;
  
  if(list == NULL || MAC_addr == NULL)
    return NULL;
 
  /* allocate new MACaddr node */

  MACAddr_ptr = (MACADDR *)malloc(sizeof(MACADDR));
  memcpy(MACAddr_ptr->MAC_addr, MAC_addr, 6);
  MACAddr_ptr->next = NULL;
  MACAddr_ptr->prev =NULL;
  
  if(list->head == NULL && list->tail == NULL){
    list->head = MACAddr_ptr;
    list->tail = MACAddr_ptr;
  }
  else{
    MACAddr_ptr->prev = list->tail;
    list->tail->next = MACAddr_ptr;
    list->tail = MACAddr_ptr;
  }
  return MACAddr_ptr;
}


/*
 * Function: int MACSpoofAddrCompareFunc(ubi_trItemPtrItemPtr,
 *                                               ubi_trNodePtr NodePtr)
 *
 * Purpose: Callback function used by splay trees to sort deauthstation node
 *
 * Arguments:
 *
 * Returns:
 *
 */
int MACSpoofedAddrCompareFunc(ubi_trItemPtr ItemPtr, ubi_trNodePtr NodePtr)
{
  Spoofed_MACaddr *A = (Spoofed_MACaddr *) NodePtr;
  Spoofed_MACaddr *B = (Spoofed_MACaddr *) ItemPtr;
  int32_t memcmp_result;
  
  memcmp_result = memcmp(&A->addr[0], &B->addr[0], 6);
  if(memcmp_result < 0){
    return 1;
  }
  else if(memcmp_result > 0){
    return -1;
  }
  return 0;
}

  
/*
 * Function: void AddMACSpoofedAddr(u_int32_t wst_addr)
 *
 * Purpose: Adds a new spoofed addr into the spoofed MAC addr tree
 *
 * Arguments: wst_addr => memory address where MAC address is recorded inside wst.
 * 
 * Returns: none
 *
 */
void AddMACSpoofedAddr(u_int32_t wst_addr)
{
  Spoofed_MACaddr *ms = NULL;
  MemBucket *mb = NULL;
  
  /* borrow a MAC address node from the MAC addresses node pool */
  
  mb = mempool_alloc(&MACspoof_data.spoofed_MACaddrPool);
  
  if(mb == NULL){
    DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF, "out of MACSpoofed addr Nodes\n"););
    return;
  }
  
  ms = (Spoofed_MACaddr *) mb->data;
  ms->bucket = mb;
  
  /* fill in the Spoofed_MACaddr struct */
  
  ms->wst_addr = wst_addr;
  memcpy(&ms->addr[0], (u_int8_t *) wst_addr, 6);
  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			  "Assigning a MACSpoofed addr of"
			  " %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
			  ms->addr[0],ms->addr[1],ms->addr[2],ms->addr[3],ms->addr[4],ms->addr[5]););
  
  /* add this spoofed MAC address to the spoofed MAC addresses tree */
  
  if(ubi_sptInsert(MACspoof_data.spoofed_MACaddrsPtr,
		   (ubi_btNodePtr)ms,
		   (ubi_btNodePtr)ms, NULL) == ubi_trFALSE){
    DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			    "Insert into MACSpoofed addr tree failed\n"););
  }
  else{
    DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			    "Insert into MACSpoofed addr tree successful\n"););
  }
  return;
}


/*
 * Function: u_int32_t PruneMACSPoofedAddrs(u_int32_t now)
 *
 * Purpose: Checks whether or not MAC addresses inserted inside spoofed
 * MAC addresses tree are still spoofed or not.
 * If not, removes concerned addresses from spoofed MAC addresses tree and
 * calls MACSpoofLogs function.
 *
 * Arguments: now => packet's capture time in sec
 *
 * Returns: number of MAC addresses removed from spoofed MAC addresses memory
 * pool.
 *
 */
u_int32_t PruneMACSpoofedAddrs(u_int32_t now)
{
  Spoofed_MACaddr *idx;
  u_int32_t pruned = 0;
  u_int32_t ms_sec;
  u_int8_t *alert_flag_ptr;

  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			  "Prune MACSpoofed addr called now: "
			  " %u , count: %u\n",
			  now, ubi_trCount(MACspoof_data.spoofed_MACaddrsPtr)););
  
  if(ubi_trCount(MACspoof_data.spoofed_MACaddrsPtr) < 1){
    DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
			    "exit PruneMACSpoofedAddrs function since MACSpoofed addr tree is free\n"););
    return(0);
  }    
  
  /* number of thing that need to be deleted */

  idx = (Spoofed_MACaddr *) ubi_btFirst((ubi_btNodePtr)MACspoof_data.spoofed_MACaddrsPtr->root);
  
  if(idx == NULL){
    return(0);
  }
  
  
  do{
    ms_sec = *((u_int32_t *) (idx->wst_addr + MACspoof_data.offset_time));
    
    if((ms_sec + MACspoof_data.expire_timeout) < now){
      
      /* MAC addresses shouldn't much more be considered as a spoofed one.
	 Removes it from spoofed MAC adresses pool */
      
      alert_flag_ptr = (u_int8_t *) (idx->wst_addr + MACspoof_data.offset_alert_flag);
      *alert_flag_ptr = 0;
      
      MACSpoofLog(NULL, idx->wst_addr);
      Spoofed_MACaddr *savidx = idx;
      if(ubi_trCount(MACspoof_data.spoofed_MACaddrsPtr) > 1){
	idx = (Spoofed_MACaddr *) ubi_btNext((ubi_btNodePtr)idx);
	DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF,
				"pruning stale MACSpoofed addr\n"););
	DeleteMACSpoofedAddr(savidx);
	pruned++;
      }
      else{
	DeleteMACSpoofedAddr(savidx);
	pruned++;
	return(pruned);
      }
    }
    else{
      if(idx != NULL && ubi_trCount(MACspoof_data.spoofed_MACaddrsPtr)){
	idx = (Spoofed_MACaddr *) ubi_btNext((ubi_btNodePtr)idx);
      }
      else{
	return pruned;
      }
    }
  } while(idx != NULL);
  return pruned;
}


/*
 * Function: void DeleteMACSpoofedAddr(Spoofed_MACaddr  *ms)
 *
 * Purpose: removed MAC address from spoofed MAC addresses mempool
 *
 * Arguments: ms => points to concerned MAC address
 *
 * Returns: none
 *
 */
void DeleteMACSpoofedAddr(Spoofed_MACaddr  *ms)
{
  Spoofed_MACaddr *oldms;
  oldms = (Spoofed_MACaddr *) ubi_sptRemove(MACspoof_data.spoofed_MACaddrsPtr,
				   (ubi_btNodePtr) ms);
  mempool_free(&MACspoof_data.spoofed_MACaddrPool, ms->bucket);
}


/*
 * Function: void MACSpoofLog(Packet *p, u_int32_t wst_addr)
 *
 * Purpose: Called either when MAC spoofing is detected or finished.
 * Logs data to a text file: MACspoof_data.logfile.
 * 
 * Arguments: p => pointer to current packet
 *            wst_addr => memory address where MAC address is recorded inside wst.
 *
 * Returns: none
 *
 */
void MACSpoofLog(Packet *p, u_int32_t wst_addr)
{
  char timestamp[TIMEBUF_SIZE];
  
  if(p){
    ts_print((struct timeval *) (wst_addr + MACspoof_data.offset_time), timestamp);
    fprintf(MACspoof_data.logfile, "\n%s\nabnormal sequence number gap reported on addr :"
	    " %.2x:%.2x:%.2x:%.2x:%.2x:%.2x, sed number: %u\n",  
	    timestamp,
	    *((u_int8_t *)(wst_addr)), *((u_int8_t *)(wst_addr + 1)), *((u_int8_t *)(wst_addr + 2)),
	    *((u_int8_t *)(wst_addr + 3)), *((u_int8_t *)(wst_addr + 4)), *((u_int8_t *)(wst_addr + 5)),
	    p->wifih->seqnum);
    
    PrintWifiHeader(MACspoof_data.logfile, p);
    
    if(p->iph != NULL){
      PrintIPHeader(MACspoof_data.logfile, p);
      
      switch(p->iph->ip_proto)
	{
	case IPPROTO_TCP:
	  PrintTCPHeader(MACspoof_data.logfile, p);
	  break;
	  
	case IPPROTO_UDP:
	  PrintUDPHeader(MACspoof_data.logfile, p);
	  break;
	  
	case IPPROTO_ICMP:
	  PrintICMPHeader(MACspoof_data.logfile, p);
	  break;
	  
	default:
	  break;
	}
    }
  }
  else{
    ts_print((struct timeval *) (wst_addr + MACspoof_data.offset_time), timestamp);
    fprintf(MACspoof_data.logfile, "\n%s End abnormal sequence number gap reported on addr: "
	    "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
	    timestamp,
	    *((u_int8_t *)(wst_addr)), *((u_int8_t *)(wst_addr + 1)), *((u_int8_t *)(wst_addr + 2)),
	    *((u_int8_t *)(wst_addr + 3)), *((u_int8_t *)(wst_addr + 4)), *((u_int8_t *)(wst_addr + 5)));
  }
  
  fflush(MACspoof_data.logfile);
  return;
}


/*
 * Function: void MACSpoofCleanExitFunction(int, void *)
 *
 * Purpose:
 * 
 * Arguments: 
 *
 * Returns:
 */
void MACSpoofCleanExitFunction(int signal, void *arg)
{
  MACSpoofData *MACsd_ptr = (MACSpoofData *) arg;
  
  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF, "MACSpoofCleanExitFunction\n"););
  
  if(fclose(MACsd_ptr->logfile))
    LogMessage("%s: fclose function failed while snort exiting: %s\n",
	       MODNAME, strerror(errno));
  return;
}


/*
 * Function: void MACSpoofCleanRestartFunction(int, void *)
 *
 * Purpose:
 * 
 * Arguments: 
 *
 * Returns:
 */
void MACSpoofRestartFunction(int signal, void *arg)
{
  MACSpoofData *MACsd_ptr = (MACSpoofData *) arg;
  
  DEBUG_WRAP(DebugMessage(DEBUG_MACSPOOF, "MACSpoofRestartFunction\n"););
  
  if(fclose(MACsd_ptr->logfile))
    LogMessage("%s: fclose function failed while snort restarting: %s\n",
	       MODNAME, strerror(errno));
  return;
}

