/*
** Copyright (C) 2003 Andrew Lockhart <andrew@ingsoc.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: spp_rogue_ap.c,v 1.2 2005/06/25 05:46:21 andrew Exp $ */

/* spp_rogue_ap
 * 
 * Purpose:
 *
 * Detects unauthorized wireless APs and AdHoc networks 
 *
 * Arguments:
 *
 * <bssid>, <channel>, scan_flag [0 | 1], scan_timeout <int>, expire_timeout <int>
 *
 * Where:
 *
 * <bssid> => a single BSSID or a list specified as [<bssid1>, <bssid2> ...]
 *
 * <channel> => a single channel or a list specified as [<channel1>, <channel2>, ...]
 *
 * scan_flag => enable scanning of multiple channels
 *
 * scan_timeout => seconds to wait between scanning channels
 *
 * expire_timeout => expiration timeout for detected access points
 *
 *
 * Effect:
 *
 * Sends an alert for each rogue network it detects of the form:
 * "Rogue AP detected (BSSID XX:XX:XX:XX:XX:XX)"
 *
 * Comments:
 *
 * Detection on multiple channels will be implemented later
 *
 */

#include "generators.h"
#include "log.h"
#include "detect.h"
#include "decode.h"
#include "event.h"
#include "plugbase.h"
#include "parser.h"
#include "mstring.h"
#include "debug.h"
#include "util.h"

#ifndef __SPP_ROGUE_AP_C__
#define __SPP_ROGUE_AP_C__
#include "spp_rogue_ap.h"
#endif


#define MODNAME "spp_rogue_ap"

/* external globals from rules.c */
extern char *file_name;
extern int file_line;

RogueApData rogue_ap_data;		  /* preprocessor data */

/*
 * Function: SetupRogueAp()
 *
 * Purpose: Registers the preprocessor keyword and initialization 
 *          function into the preprocessor list.  This is the function that
 *          gets called from InitPreprocessors() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void SetupRogueAp()
{
  /* link the preprocessor keyword to the init function in 
	  the preproc list */
  RegisterPreprocessor("rogue_ap", RogueApInit);
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: RogueAp is setup...\n"););
}


/*
 * Function: RogueApInit(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void RogueApInit(u_char *args)
{
    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Preprocessor: RogueAp Initialized\n"););

    /* parse the argument list from the rules file */
    RogueApParseArgs(args);

    /* Set the preprocessor function into the function list */
    AddFuncToPreprocList(RogueApDetect);
/*     AddFuncToCleanExitList(RogueApCleanExit); */
/*     AddFuncToRestartList(RogueApRestart); */


    if(rogue_ap_data.scan_flag){
#warning "Implement channel hopping"
    }
}



/*
 * Function: RogueApParseArgs(char *)
 *
 * Purpose: Process the preprocessor arguements from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */
  
void RogueApParseArgs(char *args)
{
  char *end_ptr;
  char **arg_toks = NULL, **arg_args = NULL;
  int i, n, n_args, n_arg_args;
  BSSIDList *bssid_list;
  u_int8_t bssid[6] = { 0 };
  
  
  memset(&rogue_ap_data, 0x00, sizeof(rogue_ap_data));
  rogue_ap_data.scan_timeout = ROGUE_AP_SCAN_TIMEOUT;	/* default scan_timeout */
  rogue_ap_data.expire_timeout = ROGUE_AP_EXPIRE_TIMEOUT; /* default expire_timeout */

  if(args == NULL)
	 return;

  rogue_ap_data.pass_list = BuildApList(&args);
  rogue_ap_data.rogue_list = AllocBSSIDList();
  rogue_ap_data.channel_mask = BuildChannelMask(&args);
  
  if(*args){
    arg_toks = mSplit(args, ",", 3, &n_args, 0);
    
	 if(n_args != 3){
		FatalError("%s: %s (%d) => rogue_ap configuration format: <bssid>, <channel>,"
					" scan_flag [ 0 | 1 ], scan_timeout <num>, expire_timeout <num>\n", MODNAME, file_name, file_line);
	 }

	 for(i = 0; i < n_args; i++){
      arg_args = mSplit(arg_toks[i], " ", 2, &n_arg_args, 0);
      if(n_arg_args != 2){
		  FatalError("%s: %s (%d) => rogue_ap configuration format: <bssid>, <channel>,"
						 " scan_flag [ 0 | 1 ], scan_timeout <num>, expire_timeout <num>\n", MODNAME, file_name, file_line);
      }
      
      if(strcmp(arg_args[0], "scan_flag") == 0){
        n = strtol(arg_args[1], NULL, 10);
        if(n != 0 && n != 1){
			 FatalError("%s: %s (%d) => \"scan_flag\" must either be 0 or 1\n", MODNAME, file_name, file_line);
        }
        rogue_ap_data.scan_flag = n;
      }
      else if(strcmp(arg_args[0], "scan_timeout") == 0){
        n = strtol(arg_args[1], NULL, 10);
        if(n <= 0){
          FatalError("%s: %s (%d) => \"scan_timeout\" must be non-negative\n", MODNAME, file_name, file_line);
        }
        rogue_ap_data.scan_timeout = n;
      }
      else if(strcmp(arg_args[0], "expire_timeout") == 0){
        n = strtol(arg_args[1], NULL, 10);
        if(n <= 0){
          FatalError("%s: %s (%d) => \"expire_timeout\" must be non-negative\n", MODNAME, file_name, file_line);
        }  
        rogue_ap_data.expire_timeout = n;
      }
		else{
		  FatalError("%s: %s (%d) => Invalid argument \"%s %s\"\n", MODNAME, file_name, file_line, arg_args[0], arg_args[1]);
		}
    }
  }
  
}



/*
 * Function: RogueApDetect(Packet *)
 *
 * Purpose: Detect APs and AdHoc networks excluding ones specified
 *          by the ACCESS_POINTS variable in snort.conf
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */

void RogueApDetect(Packet *p)
{
  Event event;
  char log_message[128];
  BSSID *bssid_ptr;
  u_int8_t *bssid;
  time_t time_curr;
  
  if(!p->wifih)
    return;
  
  time_curr = time(NULL);
  
  switch(p->wifih->type){		  /* point bssid to the correct field in the frame */

  case TYPE_MANAGEMENT:
	 bssid = p->wifih->addr3;
	 break;

  case TYPE_CONTROL:

	 if(p->wifih->stype == STYPE_RTS || p->wifih->stype == STYPE_CTS || p->wifih->stype == STYPE_ACK){
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Received Control frame subtype that does not contain BSSID\n", MODNAME););
		return;
	 }
	 else if(p->wifih->stype == STYPE_PS)
		bssid = p->wifih->addr1;
	 else if(p->wifih->stype == STYPE_CFEND)
		bssid = p->wifih->addr2;
	 else if(p->wifih->stype == STYPE_CFEND_CFACK)
		bssid = p->wifih->addr2;
	 else{
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Received Control frame with unknown/invalid subtype\n", MODNAME););
		return;
	 }
	 
	 break;
	 
  case TYPE_DATA:
	 if(p->wifih->to_ds){		
		if(p->wifih->from_ds){	  /* to_ds == 1 && from_ds == 1 */
		  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Received Data frame subtype that does not contain BSSID\n", MODNAME););
		  return;
		}
		else							  /* to_ds == 1 && from_ds == 0 */
		  bssid = p->wifih->addr1;
	 }
	 else{
		if(p->wifih->from_ds)	  /* to_ds == 0 && from_ds == 1 */
		  bssid = p->wifih->addr2;
		else							  /* to_ds == 0 && from_ds == 0 */
		  bssid = p->wifih->addr3;
	 }
		
	 break;
	 
  default:
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Received frame with unknown/invalid type\n", MODNAME););
	 return;
  }
  
  
  /* iterate through rogue AP list */
  bssid_ptr = rogue_ap_data.rogue_list->head;
  while(bssid_ptr){
	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Checking %.2X:%.2X:%.2X:%.2X:%.2X:%.2X against rogue "
									 "(BSSID %.2X:%.2X:%.2X:%.2X:%.2X:%.2X)\n", MODNAME, 
									 bssid[0], bssid[1], bssid[2],
									 bssid[3], bssid[4], bssid[5],
									 bssid_ptr->bssid[0], bssid_ptr->bssid[1], bssid_ptr->bssid[2], 
									 bssid_ptr->bssid[3], bssid_ptr->bssid[4], bssid_ptr->bssid[5]););
	 
	 if(time_curr - bssid_ptr->timestamp >= rogue_ap_data.expire_timeout){
		
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Removing expired AP from rogue AP list (BSSID %.2X:%.2X:%.2X:%.2X:%.2X:%.2X)\n",
										MODNAME, bssid_ptr->bssid[0], bssid_ptr->bssid[1], bssid_ptr->bssid[2], 
										bssid_ptr->bssid[3], bssid_ptr->bssid[4], bssid_ptr->bssid[5]););
		
		bssid_ptr = RemoveBSSID(bssid_ptr, rogue_ap_data.rogue_list);	/* remove BSSIDs that have expired */
		continue;
	 }
   
  	 if(memcmp(bssid, bssid_ptr->bssid, 6) == 0){
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: BSSID already on rogue AP list (BSSID %.2X:%.2X:%.2X:%.2X:%.2X:%.2X)\n",
										MODNAME, bssid_ptr->bssid[0], bssid_ptr->bssid[1], bssid_ptr->bssid[2], 
										bssid_ptr->bssid[3], bssid_ptr->bssid[4], bssid_ptr->bssid[5]););
		return;						  /* return if BSSID is already on the list */ 
	 }
	 
	 bssid_ptr = bssid_ptr->next;
  }
  
  /* iterate through legitimate AP list */
  for(bssid_ptr = rogue_ap_data.pass_list->head; bssid_ptr != NULL; bssid_ptr = bssid_ptr->next)
    if(memcmp(bssid_ptr->bssid, bssid, 6) == 0){
		DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: BSSID on good AP list (BSSID %.2X:%.2X:%.2X:%.2X:%.2X:%.2X)\n",
										MODNAME, bssid_ptr->bssid[0], bssid_ptr->bssid[1], bssid_ptr->bssid[2], 
										bssid_ptr->bssid[3], bssid_ptr->bssid[4], bssid_ptr->bssid[5]););
      return;						  /* return if the AP is legit */
    }
  /* construct alert message */
  snprintf(log_message, 128, "Rogue AP detected (BSSID %.2X:%.2X:%.2X:%.2X:%.2X:%.2X)",
			  bssid[0], bssid[1], bssid[2],
			  bssid[3], bssid[4], bssid[5]);

  SetEvent(&event, GENERATOR_SPP_ROGUE_AP, ROGUE_AP_DETECTED, 1, 4, 2, 0);
/*   if(p->iph == NULL)				 */  /* horrible hack to get this to show up in ACID */
/* 	 p->iph = (IPHdr *)calloc(1, sizeof(IPHdr)); */
  
  CallAlertFuncs(p, log_message, NULL, &event);	/* send alert */
  
  InsertBSSID(bssid, rogue_ap_data.rogue_list); /* insert BSSID into rogue AP list */
  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to rogue AP list\n", MODNAME,
								  bssid[0], bssid[1], bssid[2],
								  bssid[3], bssid[4], bssid[5]););
}



void RogueApCleanExitFunction(int signal)
{
#warning "Implement RogueApCleanExitFunction()"
       /* clean exit code goes here */
}



void RogueApRestartFunction(int signal)
{
#warning "Implement RogueApRestartFunction()"
       /* restart code goes here */
}



/*
 * Function: BuildApList(char **)
 *
 * Purpose: Parses bssid argument from the preprocessor's arguments string into a 
 *          BSSIDList and points it to the end of the bssid argument
 *
 * Arguments: args_ptr => pointer to a string containing the preprocessor's arguments
 *
 * Returns: a BSSIDList of BSSIDs to exclude from alerts
 *
 */

BSSIDList *BuildApList(char **args_ptr){
  char *end_ptr;
  char *args = *args_ptr;
  BSSIDList *bssid_list = NULL;
  u_int8_t bssid[6];

  if(args == NULL || *args == '\0')
	 FatalError("%s: %s (%d) => rogue_ap configuration format: <bssid>, <channel>,"
					" scan_flag [ 0 | 1 ], scan_timeout <num>, expire_timeout <num>\n", MODNAME, file_name, file_line);
  

  while(*args && isspace(*args))						  /* eat up initial whitespace */
	 args++;

  bssid_list = AllocBSSIDList();
  
  memset(bssid, 0xff, 6);
  InsertBSSID(bssid, bssid_list);

  /* parse the bssid or list of bssids */
  if(args[0] == '['){
    
    for(end_ptr = ++args; *end_ptr && *end_ptr != ']'; end_ptr++)
      if(*end_ptr == ','){
        *end_ptr = '\0';
        if(ParseBSSID(bssid, args) < 0)
          FatalError("%s: %s (%d) => Error parsing BSSID \"%s\"\n",
                     MODNAME, file_name, file_line, args);

		  DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to good AP list\n", 
										  MODNAME, bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]););
        InsertBSSID(bssid, bssid_list);
		  args = end_ptr + 1;
      }
    
    *end_ptr = '\0';
    if(ParseBSSID(bssid, args) < 0)
      FatalError("%s: %s (%d) => Error parsing BSSID \"%s\"\n",
                 MODNAME, file_name, file_line, args);

	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to good AP list\n", 
									 MODNAME, bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]););
	 InsertBSSID(bssid, bssid_list);
    args = end_ptr + 1;
	 while(*args && *args != ',') args++;
	 if(*args) args++;
  }
  else{
    
    for(end_ptr = args; *end_ptr && *end_ptr != ','; end_ptr++);
    
    *end_ptr = '\0';
    if(ParseBSSID(bssid, args) < 0)
      FatalError("%s: %s (%d) => Error parsing BSSID \"%s\"\n",
                 MODNAME, file_name, file_line, args);

	 DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN, "%s: Adding %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to good AP list\n", 
									 MODNAME, bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]););
    InsertBSSID(bssid, bssid_list);
    args = end_ptr + 1;
  }
  
  *args_ptr = args;

  return bssid_list;
}



/*
 * Function: BuildChannelMask(char **)
 *
 * Purpose: Parses channel argument from the preprocessor's arguments string into a 
 *          bitmask and points it to the end of the channel argument
 *
 * Arguments: args_ptr => pointer to a string containing the preprocessor's arguments
 *
 * Returns: a 16 bit mask containing the channels to be monitored
 *
 */

u_int16_t BuildChannelMask(char **args_ptr){
  char *end_ptr;
  char *args = *args_ptr;
  u_int16_t mask = 0;
  int n;
  
  if(args == NULL || *args == '\0')
	 FatalError("%s: %s (%d) => rogue_ap configuration format: <bssid>, <channel>,"
					" scan_flag [ 0 | 1 ], scan_timeout <num>, expire_timeout <num>\n", MODNAME, file_name, file_line);
    
  while(*args && isspace(*args)) args++; /* eat initial white space */
											  
  
  if(args[0] == '['){
    for(end_ptr = ++args; *end_ptr && *end_ptr != ']'; end_ptr++)
      if(*end_ptr == ','){
        *end_ptr = '\0';
        n = strtol(args, NULL, 10);
        if(n < ROGUE_AP_MIN_CHANNEL || n > ROGUE_AP_MAX_CHANNEL)
          FatalError("%s: %s (%d) => Channel \"%s\" out of range\n",
                     MODNAME, file_name, file_line, args);
        mask |= 1 << n;
        args = end_ptr + 1;
      }
    
    *end_ptr = '\0';
    n = strtol(args, NULL, 10);
    if(n < ROGUE_AP_MIN_CHANNEL || n > ROGUE_AP_MAX_CHANNEL)
      FatalError("%s: %s (%d) => Channel \"%s\" out of range\n",
                 MODNAME, file_name, file_line, args);
    mask |= 1 << n;
    args = end_ptr + 1;    
 
	 /* eat up to and including next ',' */
	 while(*args && *args != ',') args++;
	 if(*args) args++;
  }
  else{
    for(end_ptr = args; *end_ptr && *end_ptr != ','; end_ptr++);
    *end_ptr = '\0';
    n = strtol(args, NULL, 10);
    if(n < ROGUE_AP_MIN_CHANNEL || n > ROGUE_AP_MAX_CHANNEL)
      FatalError("%s: %s (%d) => Channel \"%s\" out of range\n",
                 MODNAME, file_name, file_line, args);
    mask |= 1 << n;
    args = end_ptr + 1;
  }
  

  
  *args_ptr = args;
  
  return mask;
}



/*
 * Function: AllocBSSIDList(void)
 *
 * Purpose: Allocates a BSSIDList
 *
 * Arguments: None
 *
 * Returns: A pointer to the newly allocated BSSIDList or
 *          NULL on error
 *
 */


BSSIDList *AllocBSSIDList(void){
  return (BSSIDList *)calloc(1, sizeof(BSSIDList));
}



void DestroyBSSIDList(BSSIDList *list){
#warning "Implement DestroyBSSIDList()"
}



/*
 * Function: InsertBSSID(u_int8_t *, BSSIDList *)
 *
 * Purpose: Creates a BSSID node from a byte array and inserts
 *          it into the specified BSSIDList
 *
 * Arguments: bssid => byte array containing a BSSID
 *            list => BSSIDList to insert new node in
 *
 * Returns: pointer to newly inserted BSSID node or NULL
 *          if bssid or list is NULL
 *
 */

BSSID *InsertBSSID(u_int8_t *bssid, BSSIDList *list){
  BSSID *bssid_ptr = NULL;

  if(list == NULL || bssid == NULL)
	 return NULL;
  
  /* allocate new BSSID node */
  bssid_ptr = (BSSID *)malloc(sizeof(BSSID));
  bssid_ptr->not_flag = 0;
  memcpy(bssid_ptr->bssid, bssid, 6);
  bssid_ptr->timestamp = time(NULL);
  bssid_ptr->next = NULL;
  bssid_ptr->prev = NULL;
  
  if(list->head == NULL && list->tail == NULL){	/* if empty list, set head and tail */
	 list->head = bssid_ptr;
	 list->tail = bssid_ptr;
  }
  else{								  /* if already a tail, link new node in */
	 bssid_ptr->prev = list->tail;
	 list->tail->next = bssid_ptr;
	 list->tail = bssid_ptr;
  }
  
  return bssid_ptr;				  /* return pointer to new BSSID node */
}



/*
 * Function: RemoveBSSID(BSSID *, BSSIDList *)
 *
 * Purpose: Removes a BSSID node from a BSSIDList
 *
 * Arguments: node => pointer to BSSID node to be removed
 *            list => BSSIDList to remove node from
 *
 * Returns: next BSSID node in the list or NULL if one
 *          does not exist
 */

BSSID *RemoveBSSID(BSSID *node, BSSIDList *list){
  BSSID *bssid_ptr;
  
  if(list == NULL)
	 return node;
  
  if(node == list->head){
	 if((list->head = node->next) == NULL)
		list->tail = NULL;
	 else
		node->next->prev = NULL;
  }
  else{
	 if((node->next->prev = node->next) == NULL)
		list->tail = node->prev;
	 else
		node->next->prev = node->prev;
  }
  
  bssid_ptr = node->next;
  free(node);
  
  return bssid_ptr;
}



/*
 * Function: ParseBSSID(u_int8_t *, char *)
 *
 * Purpose: Parses BSSIDs from a character string and stores them
 *          in a byte array.
 *
 * Arguments: bssid => pointer to byte array to store parsed BSSID in
 *            mac_str => character string containing BSSID
 *
 * Returns: 0 on success, -1 on error
 *
 */

int ParseBSSID(u_int8_t *bssid, char *mac_str){
  int i = 0;
  char *next = mac_str;

  while(*next){
    if((bssid[i] = strtol(mac_str, &next, 16)) > 255){
      return -1;
    }
    if(mac_str != next && i < 6){
      mac_str = next + 1;
      i++;
    }
    else
      return -1;
  }

  return 0;
}
