/*
** Copyright (C) 2004 Sebastien Gracia <wireless@sgdn.pm.gouv.fr>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** Copyright disclaimers:
** The Directeur Central de la scurit des systmes d'information hereby
** disclaims all copyright interest in the changes and enhancements
** made by sebastien Gracia.
** 5th Fev 2004.  
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#ifdef HAVE_MAP_ANONYMOUS
#include <sys/mman.h>
#endif

#include "decode.h"
#include "util.h"
#include "debug.h"
#include "generators.h"
#include "log.h"
#include "plugbase.h"
#include "detect.h"
#include "snort.h"

#ifndef __WIFI_STATETABLE_C__
#define __WIFI_STATETABLE_C__
#include "wifi_statetable.h"
#endif

static u_int16_t current_size = WST_DATA_INITIAL_LENGTH;
/* Total size of data associated with a MAC address inside the WST
 * (Wifi StateTable). Incremented during preprocessors initialization
 * whenever a preprocessor request new data recording */

static u_int8_t *WST_ptr;  /* pointer to the WST */

static u_int32_t rdm = 0;  /* random value */
/* MAC address and corresponding data will be inserted inside WST through
 * a Hashing process (see WSTSearchAndInsert comments). rdm is used to avoid
 * predictable Hash value */

static u_int32_t alert_time = 0;
/* time in second when the last "RANDOM_MAC_ADDR_TRAFFIC" alert was sended */

static u_int32_t timeout = 0;
/* Time validity of a MAC address and associated data recorded inside the 
 * WST. WSTSearchAndInsert function considers as decayed a MAC address and data
 * with a timeval older than timeout.
 * Update throught WSTUpdateTimeout. */

/*
 * Function: u_int16_t WSTInit(u_int8_t size)
 *
 * Purpose: Called by WST's preprocessors.
 * Updates current_size and returns data offset corresponding to
 * calling preprocessor's needs.
 *
 * Arguments: size => data's size
 *
 * Returns: data's offset
 */
u_int16_t WSTInit(u_int8_t size)
{
  u_int16_t buffer_size;

  buffer_size = current_size;
  current_size += size;
  
  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "WSTInit function."
			  "current_size = %u, was just %u bytes increased\n",
			  current_size, size););
  
  return(buffer_size);
}


/*
 * Function: void WSTDoTable(void)
 *
 * Purpose: Creates WST called by SnortMain.
 *
 * Arguments: None
 *
 * Returns: void
 */
void WSTDoTable(void)
{
  static u_int32_t wst_total_size;  /* total size in bytes of WST */
  
  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "Creating WST\n"););

  if(current_size == WST_DATA_INITIAL_LENGTH){
    if(!pv.quiet_flag){
      LogMessage("No memory allocated since no preprocessor requires it");
      LogMessage("\n+++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    }
    return;
  }
  
  srand(time(0));
  rdm = rand();  /* allocating a radom value to rdm */
  
  wst_total_size = current_size*WST_LENGTH;

#ifdef HAVE_MAP_ANONYMOUS
  WST_ptr = mmap(NULL, wst_total_size, PROT_READ | PROT_WRITE ,
		 MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
  if (WST_ptr == MAP_FAILED){
    FatalError("WSTDoTable failed to map requested length into memory: %s\n",
	       strerror(errno));
  }
#else
  WST_ptr = malloc(wst_total_size);
  if (WST_ptr == NULL){ 
    FatalError("WSTDoTable failed to allocate requested memory: %s\n",
	       strerror(errno));
  }
#endif
  if(!pv.quiet_flag){
    LogMessage("%u Kbyte allocated\n", (wst_total_size >> 10));
    LogMessage("    %s: %u\n", OPT_WST_LENGTH, WST_LENGTH);
    LogMessage("    %s: %u\n", OPT_WST_TIMEOUT , timeout);
    LogMessage("    %s: %u\n", OPT_MAX_HASH_RETRY, MAX_HASH_RETRY);
    LogMessage("    %s: %u", OPT_WST_ALERT_PERIOD, WST_ALERT_PERIOD);
    LogMessage("\n+++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  }
  memset(WST_ptr , 0, wst_total_size);
  
  srand(time(0));
  rdm = rand();  /* allocating a radom value to rdm */
  
  AddFuncToCleanExitList(WSTCleanExitFunction, &wst_total_size);
  AddFuncToRestartList(WSTRestartFunction, &wst_total_size);
 
  return;
}

/*
 * Function: void WSTUpdateTimeout(u_int32_t time)
 *
 * Purpose: Called by WST preprocessors.
 * Whenever called, WSTmain checks whether or not needed timeout is superior or not
 * to actual WST timeout.
 * If not, updates it.
 *
 * Arguments: time => WST timeout length requested by calling preprocessor.
 *
 * Returns: void
 */
void WSTUpdateTimeout(u_int32_t time)
{
  if(time * 2 >= timeout){
    timeout = time * 2;
  }
  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, " WST timeout =%u\n", timeout););
  return;
}


/*
 * Function: void WSTMain(Packet *p, wst_flag_type flag_addr)
 *
 * Purpose: Called by WST preprocessors.
 * Whenever called, WSTmain checks whether or not a previous preprocessor
 * already made a request for corresponding MAC address packet (src or dest).
 * If not, fills up desired p->wstidx.cache_addr by calling WSTSearchAndUInsert
 * function.
 *
 * Arguments: p => pointer to the current packet data struct
 *
 *            flag_addr => either src or dst wst_flag_type depending on
 *            calling preprocessors needs.
 *
 * Returns: void
 */
void WSTMain(Packet *p, wst_flag_type flag_addr)
{
  u_int8_t *MACaddr_ptr; /* pointer to p->wifih corresponding to desired
			    MAC address */
  int8_t offset;
  
  /* checks whether or not p->wstidx.cache_addr[flag_addr] is already filled up. 
   * if not determine MACaddr_ptr and finally fills p->wstidx.cache_addr[flag_addr] up */
  
  if(!p->wstidx.cache_addr[flag_addr]){
    offset = WSTDetermineMACAddrOffset(p, flag_addr);
    if(offset == -1){
      p->wstidx.cache_addr[flag_addr] = WST_FAILED;
      return;
    }
    else{
      MACaddr_ptr =  ((u_int8_t*) p->wifih + offset);
      
      /* finally calls WSTSearchAndInsert function to determine
       * p->wstidx.cache_addr[flag_addr] */ 
      
      p->wstidx.cache_addr[flag_addr] = WSTSearchAndInsert(p, MACaddr_ptr, flag_addr); 
    }
  }
  return;
}
  

/*
 * Function: int8_t WSTDetermineMACAddrOffset(Packet *p, wst_flag_type flag_addr)
 *
 * Purpose: Depending on both kind of frame type and needs
 * (src or dst MAC address), determines needed offset inside p->wifih
 *
 * Arguments: p => pointer to the current packet data struct
 *
 *            flag_addr => either src or dst wst_flag_type depending on
 *            calling preprocessors needs.
 *
 * Returns: needed offset inside p->wifih or -1 if failed
 */
int8_t WSTDetermineMACAddrOffset(Packet *p, wst_flag_type flag_addr)
{
  
  switch(p->wifih->type){
    
  case TYPE_MANAGEMENT:
    if(flag_addr == WST_FLAG_SRCADDR){
      return(OFFSET_MANAGEMENT_SRCADDR);
    }
    else{
      return(OFFSET_MANAGEMENT_DESTADDR);
    }
    
  case TYPE_CONTROL:
    if(p->wifih->stype == STYPE_CTS || p->wifih->stype == STYPE_ACK){
      if(flag_addr == WST_FLAG_DESTADDR){
	return(OFFSET_CONTROL_DESTADDR);
      }
      else{
	DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,
				"received frame do not contain any source MAC address/invalid request\n"););
	return -1;
      }
    }
    else{
      if(flag_addr == WST_FLAG_SRCADDR){ 
	return(OFFSET_CONTROL_SRCADDR);
      }
      else{
	return(OFFSET_CONTROL_DESTADDR);
      }
    }
    
  case TYPE_DATA:
    if(p->wifih->from_ds){
      if(p->wifih->to_ds){  /* from_ds == 1 && to_ds == 1 */
	if(flag_addr == WST_FLAG_SRCADDR){
	  return(OFFSET_DATA_SRCADDR_FROM_TO);
	}
	else{
	  return(OFFSET_DATA_DESTADDR_FROM_TO);
	}
      }
      else{
	if(flag_addr == WST_FLAG_SRCADDR){  /* from_ds == 1 && to_ds == 0 */
	  return(OFFSET_DATA_SRCADDR_FROM);
	}
	else{
	  return(OFFSET_DATA_DESTADDR_FROM);
	}
      }
    }
    else{
      if(p->wifih->to_ds){  /* from_ds == 0 && to_ds == 1 */
	if(flag_addr == WST_FLAG_SRCADDR){
	  return(OFFSET_DATA_SRCADDR_TO);
	}
	else{
	  return(OFFSET_DATA_DESTADDR_TO);
	}
      }
      else{
	if(flag_addr == WST_FLAG_SRCADDR){  /* from_ds == 0 && to_ds == 0 */
	  return(OFFSET_DATA_SRCADDR_);
	}
	else{
	  return(OFFSET_DATA_DESTADDR_); 
	}
      }
    }
    
  default:
    DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "received frame with unknown/invalid type\n"););
    return -1;
  }
}

  
/*
 * Function: int32_t WSTSearchAndInsert(Packet *p, u_int8_t *MACaddr_ptr,
 * wst_flag_type flag_addr)
 *
 * Purpose: Searches for the MAC address pointed by MACaddr_prt inside WST.
 * In default, inserts it.
 * If neither able to find nor insert it, sends an alert (traffic with random
 * address detected).
 *
 * Arguments: p => pointer to the current packet data struct
 *
 *            MACaddr_ptr => pointer to p->wifih corresponding to desired
 *                           MAC address
 *
 *            flag_addr => either src or dst wst_flag_type depending upon
 *            calling preprocessors needs.
 * 
 * Returns: memory address corresponding to the place where MAC address and
 * related data are recorded inside WST.
 */
int32_t WSTSearchAndInsert(Packet *p, u_int8_t *MACaddr_ptr,
			   wst_flag_type flag_addr)
{
  u_int32_t hash, x[2];
  u_int32_t wst_offset_line; /* offset line inside WST table */
  u_int8_t cpt=0; /* number of retry while trying to determine where MAC adrress and
		   * related data are recorded inside WST */
  struct timeval wst_tv, tv_diff;
  Event event;
  char outstring[255], tmp[255];

  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "WSTSearchAndInsert function processing "
			  "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x MAC address\n",
			  *((u_int8_t *)(MACaddr_ptr)), *((u_int8_t *)(MACaddr_ptr + 1)),
			  *((u_int8_t *)(MACaddr_ptr + 2)), *((u_int8_t *)(MACaddr_ptr + 3)),
			  *((u_int8_t *)(MACaddr_ptr + 4)), *((u_int8_t *)(MACaddr_ptr + 5))););
  
  /* fill up x[0] and x[1]. x[0] contains the 4 first bytes of MAC address, x[1]
   * the last 2. To avoid later predictable hash value x[0] is gonna be xored with a random
   * value: rdm. */

  x[1]=0;
  memcpy(&x[0], MACaddr_ptr, SIZE_MAC_ADDR);
  x[0] ^= rdm;
  
  /* It's time to determine where MAC adrress and related data are recorded
   * inside WST */

  do{

    /* Determines Hash value */

    hash = WSTMix32(x[0], x[1]);
    hash %= WST_LENGTH;

    /* evaluates wst_offset_line */

    wst_offset_line = hash * current_size;

    /* Checks if wst_offset_line is the offset line associated with needed MAC
     * address (the one we are looking for) */

    if(!memcmp(MACaddr_ptr, WST_ptr + wst_offset_line, SIZE_MAC_ADDR)){
      DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "Find corresponding line offset"
			      " inside WST\n"););
      
      /* Updates timeval related with MAC address inside WST */
      
      memcpy(WST_ptr + wst_offset_line + SIZE_MAC_ADDR, &p->pkth->ts,
	     sizeof(struct timeval));

      /* returns memory address corresponding to MAC address and related data
	 we were looking for */
      
      return((int32_t) (WST_ptr + wst_offset_line));
    }

    /* calculated wst_offset_line isn't the offset line associated with needed
     * MAC address. Maybe MAC address associated with wst_offset_line should be
     * considered as decayed because of an expired associated timeval? */
    
    memcpy(&wst_tv , WST_ptr + wst_offset_line + SIZE_MAC_ADDR,
	   sizeof(struct timeval));
    tv_diff = WSTTimevalSubtract(p->pkth->ts, wst_tv);
    if(tv_diff.tv_sec > timeout){
      DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "Find a decayed MAC address"
			      " inside WST\n"););
      
      /* Overwrites decayed MAC address with the new one and updates timeval
       *  related with */
      
      memcpy(WST_ptr + wst_offset_line, MACaddr_ptr, SIZE_MAC_ADDR);
      memcpy(WST_ptr + wst_offset_line + SIZE_MAC_ADDR, &p->pkth->ts,
	     sizeof(struct timeval));
      
      /* refreshes WST data */
      
      bzero(WST_ptr + wst_offset_line + SIZE_MAC_ADDR + sizeof(struct timeval),
	    current_size - SIZE_MAC_ADDR - sizeof(struct timeval));
      
      /* returns memory address corresponding to the place where MAC address
       * and related data are now recorded inside WST */
      
      return((int32_t) (WST_ptr + wst_offset_line));
    }
    
    /* Though MAC address pointed by wst_offset_line isn't the one
     * we were looking for, associated timeval will be updated.
     * Such refresh is made too avoid any subsequent loss of data due to both
     * collision and unlucky timeval expiry situation.
     * On the other hand, such refresh might involv some "dead wst_offset_line
     * " offset. According to wireless network characteristics, it shouldn't
     * have any real consequences. */

    memcpy(WST_ptr + wst_offset_line + SIZE_MAC_ADDR, &p->pkth->ts,
	   sizeof(struct timeval));
    
    /* Raising x[1] to calculate a new hash value and furthermore
       a new wst_offset_line. */ 
    
    x[1] += 1;
    cpt++;
  } while(cpt <= MAX_HASH_RETRY);
  
  
  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "WSTSearchAndInsert function failed"
			  " to insert MAC address inside WST\n"););
  
  /* Checks whether or not an alert has already been sended before WST_ALERT_PERIOD
   * ended. */
  
  if(alert_time + WST_ALERT_PERIOD > p->pkth->ts.tv_sec){
    return( WST_FAILED);
  }
  
  /* an alert gonna be raised */ 
  
  alert_time = p->pkth->ts.tv_sec;
  
  if(flag_addr == WST_FLAG_SRCADDR){
    snprintf(outstring, 255, RANDOM_SRC_MAC);
  }
  else{
    snprintf(outstring, 255, RANDOM_DST_MAC);
  }
   snprintf(tmp, 255,
	    "\nAddr src: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x"
	    " -> Addr dst: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x"
	    ",\nBssid: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x, TYPE: 0x%x, STYPE: 0x%x.",
	    p->wifih->addr2[0], p->wifih->addr2[1], p->wifih->addr2[2],
	    p->wifih->addr2[3], p->wifih->addr2[4], p->wifih->addr2[5],
	    p->wifih->addr1[0], p->wifih->addr1[1], p->wifih->addr1[2],
	    p->wifih->addr1[3], p->wifih->addr1[4], p->wifih->addr1[5],
	    p->wifih->addr3[0], p->wifih->addr3[1], p->wifih->addr3[2],
	    p->wifih->addr3[3], p->wifih->addr3[4], p->wifih->addr3[5],
	    p->wifih->type, p->wifih->stype);
   strncat(outstring, tmp, 255);
   
   DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "%s\n", outstring););
   
   SetEvent(&event, GENERATOR_WIFI_STATETABLE, RANDOM_MAC_ADDR_TRAFFIC,
	    1, 0, 0, 0);
   CallAlertFuncs(p, outstring, NULL, &event);	/* send alert */
   
   return(WST_FAILED);
}


/*
 * Function: u_int32_t WSTMix32(u_int32_t x, u_int32_t y)
 *
 * Purpose: Calculate an hash value based on two u_int32_t. 
 *
 * Arguments: x => first u_int32_t
 *            y => second u_int32_t
 *
 * Returns: hash value
 */
u_int32_t WSTMix32(u_int32_t x, u_int32_t y)
{
  u_int32_t db5=0xfb5cc520, w31_27;
  
  w31_27 = (y>>27); y<<=5; y+=((db5>>w31_27)^(db5<<(32-w31_27)));
  w31_27 = (y>>27); y<<=5; y+=((db5>>w31_27)^(db5<<(32-w31_27)));
  
  w31_27 = (x>>27); x<<=5; x+=((db5>>w31_27)^(db5<<(32-w31_27)));

  x^=y;

 w31_27 = (x>>27); x<<=5; x+=((db5>>w31_27)^(db5<<(32-w31_27)));
 w31_27 = (x>>27); x<<=5; x+=((db5>>w31_27)^(db5<<(32-w31_27)));
 w31_27 = (x>>27); x<<=5; x+=((db5>>w31_27)^(db5<<(32-w31_27)));
 w31_27 = (x>>27); x<<=5; x+=((db5>>w31_27)^(db5<<(32-w31_27)));
 w31_27 = (x>>27); x<<=5; x+=((db5>>w31_27)^(db5<<(32-w31_27)));

 return(x);
}


/*
 * Function: struct timeval WSTTimevalSubtract(struct timeval x, struct timeval y)
 *
 * Purpose: Subtract two `struct timeval.
 *
 * Arguments: x => first struct timeval value
 *            y => second struct timeval value
 *
 * Returns: result of subtract
 */
struct timeval WSTTimevalSubtract (struct timeval x, struct timeval y)
{
  struct timeval result;
  
  /* Perform the carry for the later subtraction by updating Y. */
  if (x.tv_usec < y.tv_usec){
    int nsec = (y.tv_usec - x.tv_usec) / 1000000 + 1;
    y.tv_usec -= 1000000 * nsec;
    y.tv_sec += nsec;
  }
  if (x.tv_usec - y.tv_usec > 1000000){
    int nsec = (x.tv_usec - y.tv_usec) / 1000000;
    y.tv_usec += 1000000 * nsec;
    y.tv_sec -= nsec;
  }
  
  /* Compute the time remaining to wait.
     `tv_usec' is certainly positive. */
  result.tv_sec = x.tv_sec - y.tv_sec;
  result.tv_usec = x.tv_usec - y.tv_usec;
  
  return(result);
}


/*
 * Function: void WSTCleanExitFunction(int, void *)
 *
 * Purpose: free wifi statetable while exiting snort
 * 
 * Arguments: 
 *
 * Returns: void
 */
void WSTCleanExitFunction(int signal, void *arg)
{
  u_int32_t *size = (u_int32_t *) arg;
  
  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "WSTCleanExitFunction\n"););
  
#ifdef HAVE_MAP_ANONYMOUS
  if(munmap(NULL, *size) == -1)
    LogMessage("Munmap failure: %s\n",
	       strerror(errno));
#else
  free(WST_ptr);
#endif
  return;
}


/*
 * Function: void WSTRestartFunction(int, void *)
 *
 * Purpose: free wifi statetable while restarting snort
 * 
 * Arguments: 
 *
 * Returns: void
 */
void WSTRestartFunction(int signal, void *arg)
{
  u_int32_t *size = (u_int32_t *) arg;

  DEBUG_WRAP(DebugMessage(DEBUG_WIFI_STATETABLE, "WSTRestartFunction\n"););

#ifdef HAVE_MAP_ANONYMOUS
  if(munmap(NULL, *size) == -1)
    LogMessage("Munmap failure: %s\n",
	       strerror(errno));
#else
  free(WST_ptr);
#endif
  return;
}

