/*
** Copyright (C) 2004 Sebastien Gracia <wireless@sgdn.pm.gouv.fr>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** Copyright disclaimers:
** The Directeur Central de la scurit des systmes d'information hereby
** disclaims all copyright interest in the changes and enhancements
** made by sebastien Gracia.
** 5th Fev 2004.  
*/


#ifndef __WIFI_STATETABLE_H__
#define __WIFI_STATETABLE_H__


/*  I N C L U D E S
*******************************************************************************/


/*  D E F I N E S
*******************************************************************************/

#define SIZE_MAC_ADDR (u_int32_t) 6

#define WST_DATA_INITIAL_LENGTH SIZE_MAC_ADDR + sizeof(struct timeval)
/* intitial size of data recorded inside the WST (Wifi StateTable),
 * corresponding to the size of a MAC address and a struct timeval */


#define WST_LENGTH (u_int32_t) 1 << 18
/* Total lenght of the WST, corresponding to the maximum number of valid MAC
 * address recorded inside. A valid MAC address means a MAC address with an
 * associated timeval below WST_TIMEOUT */

#define MAX_HASH_RETRY (u_int8_t) 5
/* Maximum number of retry for the WSTSearchAnsInsert function to insert or
 * search for a MAC address inside the WST. Above this limit an alert will be 
 * send because someone around sending wifi frames with numerous random either
 * source or destination MAC address */

#define WST_ALERT_PERIOD 120
/* Number of second to wait for sending a new "RANDOM_MAC_ADDR-TRAFFIC" alert
 * if such a situation is still occur */

#define WST_FAILED -1
/* value return by WSTSearchAnsInsert function whenever MAX_HASH_RETRY is
 *  overruned or WSTMain received frame with unknow/invalid type */

#define OPT_WST_LENGTH "wst_length"
#define OPT_WST_TIMEOUT "wst_timeout"
#define OPT_MAX_HASH_RETRY "wst_max_hash_retry"
#define OPT_WST_ALERT_PERIOD "wst_alert_period"

typedef enum {
  /* 802.11 management frame address offset */
  OFFSET_MANAGEMENT_SRCADDR = 0x0a,
  OFFSET_MANAGEMENT_DESTADDR = 0x04,

 /* 802.11 control frame address offset */
  OFFSET_CONTROL_SRCADDR = 0x0a,
  OFFSET_CONTROL_DESTADDR = 0x04,

 /* 802.11 data frame address offset */
  OFFSET_DATA_SRCADDR_FROM_TO = 0x0a,
  OFFSET_DATA_DESTADDR_FROM_TO = 0x10,

  OFFSET_DATA_SRCADDR_FROM = 0x10,
  OFFSET_DATA_DESTADDR_FROM = 0x04,

  OFFSET_DATA_SRCADDR_TO = 0x0a,
  OFFSET_DATA_DESTADDR_TO = 0x10,

  OFFSET_DATA_SRCADDR_ = 0x0a,
  OFFSET_DATA_DESTADDR_ = 0x04
} addr_offset_t;
/* Whenever called WSTSearchAndInsert function will record data associated on
 * either source or destination MAC address, depending on calling
 * preprocessor's needs. Thanks to addr_offset_t, WSTSearchAndInsert will look
 * for the right offset for each packet */

typedef enum {
  WST_FLAG_SRCADDR = 0x00,
  WST_FLAG_DESTADDR = 0x01
} wst_flag_type;


/* borrowed from airsnort */
typedef enum {                  /* 802.11 frame types */
  TYPE_INVALID = -1,
  TYPE_MANAGEMENT = 0x00,
  TYPE_CONTROL = 0x01,
  TYPE_DATA = 0x02
} wifi_type_t;


typedef enum {                 /* 802.11 management frame subtypes */
  STYPE_ASSOCREQ = 0x00,       /* Subtypes are already AND'ed with their */
  STYPE_ASSOCRESP = 0x01,      /* corresponding Types to prevent false  */
  STYPE_REASSOC_REQ = 0x02,    /* positives and the need for checking the type */
  STYPE_REASSOC_RESP = 0x03,   /* independently */
  STYPE_PROBEREQ = 0x04,
  STYPE_PROBERESP = 0x05,
  STYPE_BEACON = 0x08,
  STYPE_ATIM = 0x09,
  STYPE_DISASSOC = 0x0a,
  STYPE_AUTH = 0x0b,
  STYPE_DEAUTH = 0x0c,
  STYPE_INVALID = 0x0d,

/* 802.11 control frame subtypes */
  STYPE_PS = 0x0a,
  STYPE_RTS = 0x0b,
  STYPE_CTS =  0x0c,
  STYPE_ACK = 0x0d,
  STYPE_CFEND = 0x0e,
  STYPE_CFEND_CFACK = 0x0f,

/* 802.11 data frame subtypes */
  STYPE_DATA = 0x00,
  STYPE_CFACK = 0x01,
  STYPE_CFPOLL = 0x02,
  STYPE_CFACK_CFPOLL = 0x03,
  STYPE_NULLFUNC = 0x04,
  STYPE_CFACK_NULLFUNC = 0x05,
  STYPE_CFPOLL_NULLFUNC = 0x06,
  STYPE_CFACK_CFPOLL_NULLFUNC = 0x07
} wifi_stype_t;


/*  P R O T O T Y P E S
*******************************************************************************/

u_int16_t WSTInit(u_int8_t);
void WSTDoTable();
void WSTUpdateTimeout(u_int32_t);
void WSTMain(Packet *, wst_flag_type);
int8_t WSTDetermineMACAddrOffset(Packet *, wst_flag_type);
int32_t WSTSearchAndInsert(Packet *, u_int8_t *, wst_flag_type);
u_int32_t WSTMix32(u_int32_t, u_int32_t);
struct timeval WSTTimevalSubtract (struct timeval , struct timeval );
void WSTCleanExitFunction(int, void *);
void WSTRestartFunction(int, void *);


#endif  /* __WIFI_STATETABLE_H__ */
